# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for the lazy loaders."""

from __future__ import annotations

import importlib.abc
import importlib.util
import sys
import warnings
from unittest import mock
import ddt

from qiskit.exceptions import MissingOptionalLibraryError, OptionalDependencyImportWarning
from qiskit.utils import LazyImportTester, LazySubprocessTester
from test import QiskitTestCase  # pylint: disable=wrong-import-order


def available_importer(**kwargs):
    """A LazyImportTester that should succeed."""
    return LazyImportTester("site", **kwargs)


def unavailable_importer(**kwargs):
    """A LazyImportTester that should fail."""
    return LazyImportTester("_qiskit_this_module_does_not_exist_", **kwargs)


def available_process(**kwargs):
    """A LazySubprocessTester that should fail."""
    return LazySubprocessTester([sys.executable, "-c", "import sys; sys.exit(0)"], **kwargs)


def unavailable_process(**kwargs):
    """A LazySubprocessTester that should fail."""
    return LazySubprocessTester([sys.executable, "-c", "import sys; sys.exit(1)"], **kwargs)


def mock_availability_test(feature):
    """Context manager that mocks out the availability checker for a given dependency checker.  The
    context manager returns the mocked-out method."""
    # We have to be careful with what we patch because the dependency managers define `__slots__`.
    return mock.patch.object(type(feature), "_is_available", wraps=feature._is_available)


def patch_imports(mapping: dict[str, importlib.abc.Loader]):
    """Patch the import system so that the given named modules will skip the regular search system
    and instead be loaded by the given loaders.

    Already imported modules will not be affected; this should use uniquely named modules."""

    class OverrideLoaders(importlib.abc.MetaPathFinder):
        """A metapath finder that will simply return an explicit loader for specific modules."""

        def __init__(self, mapping: dict[str, importlib.abc.Loader]):
            self.mapping = mapping

        def find_spec(self, fullname, path, target=None):
            """Implementation of the abstract (but undefined) method."""
            del path, target  # ABC parameters we don't need.
            if (loader := self.mapping.get(fullname)) is not None:
                return importlib.util.spec_from_loader(fullname, loader)
            return None

    new_path = [OverrideLoaders(mapping)] + sys.meta_path
    return mock.patch.object(sys, "meta_path", new_path)


@ddt.ddt
class TestLazyDependencyTester(QiskitTestCase):
    """Tests for the lazy loaders.  Within this class, we parameterise the test cases with
    generators, rather than the mocks themselves.  That allows us to easily generate clean
    instances, and means that creation doesn't happen during test collection."""

    @ddt.data(available_importer, available_process)
    def test_evaluates_correctly_true(self, test_generator):
        """Test that the available loaders evaluate True in various Boolean contexts."""
        self.assertTrue(test_generator())
        self.assertTrue(bool(test_generator()))
        if not test_generator():
            self.fail("did not evaluate true")

    @ddt.data(unavailable_importer, unavailable_process)
    def test_evaluates_correctly_false(self, test_generator):
        """Test that the available loaders evaluate False in various Boolean contexts."""
        self.assertFalse(test_generator())
        self.assertFalse(bool(test_generator()))
        if test_generator():
            self.fail("did not evaluate false")

    def test_submodule_import_detects_false_correctly(self):
        """Test that a lazy import of a submodule where the parent is not available still generates
        a silent failure."""

        # The idea here is that the base package is what will fail the import, and the corresponding
        # `ImportError.name` won't be the same as the full path we were trying to import.  We want
        # to make sure that the "was it found and failed to import?" handling is correct in this
        # case.
        def checker():
            return LazyImportTester("_qiskit_module_does_not_exist_.submodule")

        # Just in case something else is allowing the warnings, but they should be forbidden by
        # default.
        with warnings.catch_warnings(record=True) as log:
            self.assertFalse(checker())
        self.assertEqual(log, [])

    @ddt.data(available_importer, available_process, unavailable_importer, unavailable_process)
    def test_check_occurs_once(self, test_generator):
        """Check that the test of availability is only performed once."""
        feature = test_generator()
        with mock_availability_test(feature) as check:
            check.assert_not_called()
            if feature:
                pass
            check.assert_called_once()

            if feature:
                feature.require_now("no message")
                feature.require_in_call(lambda: None)()
                feature.require_in_call("no message")(lambda: None)()
                feature.require_in_instance(type("Dummy", (), {}))()
                feature.require_in_instance("no message")(type("Dummy", (), {}))()

            check.assert_called_once()

    @ddt.data(available_importer, available_process, unavailable_importer, unavailable_process)
    def test_callback_occurs_once(self, test_generator):
        """Check that the callback is only called once."""
        callback = mock.MagicMock()

        feature = test_generator(callback=callback)

        callback.assert_not_called()
        if feature:
            pass
        callback.assert_called_once_with(bool(feature))

        callback.reset_mock()
        if feature:
            feature.require_now("no message")
            feature.require_in_call(lambda: None)()
            feature.require_in_call("no message")(lambda: None)()
            feature.require_in_instance(type("Dummy", (), {}))()
            feature.require_in_instance("no message")(type("Dummy", (), {}))()
        callback.assert_not_called()

    @ddt.data(available_importer, available_process)
    def test_require_now_silently_succeeds_for_available_tests(self, test_generator):
        """Test that the available loaders silently do nothing when they are required."""
        feature = test_generator()
        with mock_availability_test(feature) as check:
            check.assert_not_called()
            feature.require_now("no message")
            check.assert_called_once()

    @ddt.data(available_importer, available_process)
    def test_require_in_call_silently_succeeds_for_available_tests(self, test_generator):
        """Test that the available loaders silently do nothing when they are required in the
        decorator form."""
        # pylint: disable=function-redefined

        with self.subTest("direct decorator"):
            feature = test_generator()
            with mock_availability_test(feature) as check:
                check.assert_not_called()

                @feature.require_in_call
                def decorated():
                    pass

                check.assert_not_called()
                decorated()
                check.assert_called_once()

        with self.subTest("named decorator"):
            feature = test_generator()
            with mock_availability_test(feature) as check:
                check.assert_not_called()

                @feature.require_in_call("sentinel name")
                def decorated():
                    pass

                check.assert_not_called()
                decorated()
                check.assert_called_once()

    @ddt.data(available_importer, available_process)
    def test_require_in_instance_silently_succeeds_for_available_tests(self, test_generator):
        """Test that the available loaders silently do nothing when they are required in the
        decorator form."""
        # pylint: disable=function-redefined

        with self.subTest("direct decorator"):
            feature = test_generator()
            with mock_availability_test(feature) as check:
                check.assert_not_called()

                @feature.require_in_instance
                class Dummy:
                    """Dummy class."""

                check.assert_not_called()
                Dummy()
                check.assert_called_once()

        with self.subTest("named decorator"):
            feature = test_generator()
            with mock_availability_test(feature) as check:
                check.assert_not_called()

                @feature.require_in_instance("sentinel name")
                class Dummy:
                    """Dummy class."""

                check.assert_not_called()
                Dummy()
                check.assert_called_once()

    @ddt.data(unavailable_importer, unavailable_process)
    def test_require_now_raises_for_unavailable_tests(self, test_generator):
        """Test that the unavailable loaders loudly raise when they are required."""
        feature = test_generator()
        with mock_availability_test(feature) as check:
            check.assert_not_called()
            with self.assertRaisesRegex(MissingOptionalLibraryError, "sentinel message"):
                feature.require_now("sentinel message")
            check.assert_called_once()

    @ddt.data(unavailable_importer, unavailable_process)
    def test_require_in_call_raises_for_unavailable_tests(self, test_generator):
        """Test that the unavailable loaders loudly raise when the inner functions of decorators are
        called, and not before, and raise each time they are called."""
        # pylint: disable=function-redefined

        with self.subTest("direct decorator"):
            feature = test_generator()
            with mock_availability_test(feature) as check:
                check.assert_not_called()

                @feature.require_in_call
                def decorated():
                    pass

                check.assert_not_called()
                with self.assertRaisesRegex(MissingOptionalLibraryError, "decorated"):
                    decorated()
                check.assert_called_once()
                with self.assertRaisesRegex(MissingOptionalLibraryError, "decorated"):
                    decorated()
                check.assert_called_once()

        with self.subTest("named decorator"):
            feature = test_generator()
            with mock_availability_test(feature) as check:
                check.assert_not_called()

                @feature.require_in_call("sentinel message")
                def decorated():
                    pass

                check.assert_not_called()
                with self.assertRaisesRegex(MissingOptionalLibraryError, "sentinel message"):
                    decorated()
                check.assert_called_once()
                with self.assertRaisesRegex(MissingOptionalLibraryError, "sentinel message"):
                    decorated()
                check.assert_called_once()

    @ddt.data(unavailable_importer, unavailable_process)
    def test_require_in_instance_raises_for_unavailable_tests(self, test_generator):
        """Test that the unavailable loaders loudly raise when the inner classes of decorators are
        instantiated, and not before, and raise each time they are instantiated."""
        # pylint: disable=function-redefined

        with self.subTest("direct decorator"):
            feature = test_generator()
            with mock_availability_test(feature) as check:
                check.assert_not_called()

                @feature.require_in_instance
                class Dummy:
                    """Dummy class."""

                check.assert_not_called()
                with self.assertRaisesRegex(MissingOptionalLibraryError, "Dummy"):
                    Dummy()
                check.assert_called_once()
                with self.assertRaisesRegex(MissingOptionalLibraryError, "Dummy"):
                    Dummy()
                check.assert_called_once()

        with self.subTest("named decorator"):
            feature = test_generator()
            with mock_availability_test(feature) as check:
                check.assert_not_called()

                @feature.require_in_instance("sentinel message")
                class Dummy:
                    """Dummy class."""

                check.assert_not_called()
                with self.assertRaisesRegex(MissingOptionalLibraryError, "sentinel message"):
                    Dummy()
                check.assert_called_once()
                with self.assertRaisesRegex(MissingOptionalLibraryError, "sentinel message"):
                    Dummy()
                check.assert_called_once()

    def test_import_allows_multiple_modules_successful(self):
        """Check that the import tester can accept an iterable of modules."""
        # Deliberately using modules that will already be imported to avoid side effects.
        feature = LazyImportTester(["site", "sys"])
        with mock_availability_test(feature) as check:
            check.assert_not_called()
            self.assertTrue(feature)
            check.assert_called_once()

    def test_import_allows_multiple_modules_failure(self):
        """Check that the import tester can accept an iterable of modules, and will ."""
        # Deliberately using modules that will already be imported to avoid side effects.
        feature = LazyImportTester(["site", "sys", "_qiskit_module_does_not_exist_"])
        with mock_availability_test(feature) as check:
            check.assert_not_called()
            self.assertFalse(feature)
            check.assert_called_once()

    def test_import_allows_attributes_successful(self):
        """Check that the import tester can accept a dictionary mapping module names to attributes,
        and that these can be fetched."""
        name_map = {
            "_qiskit_dummy_module_1_": ("attr1", "attr2"),
            "_qiskit_dummy_module_2_": ("thing1", "thing2"),
        }
        mock_modules = {}
        for module, attributes in name_map.items():
            # We could go through the rigmarole of creating a full module with importlib, but this
            # is less complicated and should be sufficient.  Property descriptors need to be
            # attached to the class to work correctly, and then we provide an instance.
            class Module:
                """Dummy module."""

                unaccessed_attribute = mock.PropertyMock()

            for attribute in attributes:
                setattr(Module, attribute, mock.PropertyMock())
            mock_modules[module] = Module()

        feature = LazyImportTester(name_map)
        with mock.patch.dict(sys.modules, **mock_modules):
            self.assertTrue(feature)

        # Retrieve the mocks, and assert that the relevant accesses were made.
        for module, attributes in name_map.items():
            mock_module = mock_modules[module]
            for attribute in attributes:
                vars(type(mock_module))[attribute].assert_called()
            vars(type(mock_module))["unaccessed_attribute"].assert_not_called()

    def test_warns_on_import_error(self):
        """Check that the module raising an `ImportError` other than being not found is warned
        against."""

        # pylint: disable=missing-class-docstring,missing-function-docstring,abstract-method

        class RaisesImportErrorOnLoad(importlib.abc.Loader):
            def __init__(self, name):
                self.name = name

            def create_module(self, spec):
                raise ImportError("sentinel import failure", name=self.name)

            def exec_module(self, module):
                pass

        dummy = f"{__name__}_{type(self).__name__}_test_warns_on_import_error".replace(".", "_")
        tester = LazyImportTester(dummy)
        with patch_imports({dummy: RaisesImportErrorOnLoad(dummy)}):
            with self.assertWarnsRegex(
                OptionalDependencyImportWarning,
                rf"module '{dummy}' failed to import with: .*sentinel import failure.*",
            ):
                self.assertFalse(tester)

    def test_warns_on_internal_not_found_error(self):
        """Check that the module raising an `ModuleNotFoundError` for some module other than itself
        (such as a module trying to import parts of Terra that don't exist any more) is caught and
        warned against, rather than silently caught as an expected `ModuleNotFoundError`."""

        # pylint: disable=missing-class-docstring,missing-function-docstring,abstract-method

        class ImportsBadModule(importlib.abc.Loader):
            def create_module(self, spec):
                # Doesn't matter what, we just want to return any module object; we're going to
                # raise an error during "execution" of the module.
                return sys

            def exec_module(self, module):
                del module  # ABC parameter we don't care about.
                import __qiskit__some_module_that_does_not_exist

        dummy = f"{__name__}_{type(self).__name__}_test_warns_on_internal_not_found_error".replace(
            ".", "_"
        )
        tester = LazyImportTester(dummy)
        with patch_imports({dummy: ImportsBadModule()}):
            with self.assertWarnsRegex(
                OptionalDependencyImportWarning,
                rf"module '{dummy}' failed to import with: ModuleNotFoundError.*__qiskit__",
            ):
                self.assertFalse(tester)

    def test_import_allows_attributes_failure(self):
        """Check that the import tester can accept a dictionary mapping module names to attributes,
        and that these are recognized when they are missing."""
        # We can just use existing modules for this.
        name_map = {
            "sys": ("executable", "path"),
            "builtins": ("list", "_qiskit_dummy_attribute_"),
        }

        feature = LazyImportTester(name_map)
        with self.assertWarnsRegex(UserWarning, r"'builtins' imported, but attribute"):
            self.assertFalse(feature)

    def test_import_fails_with_no_modules(self):
        """Catch programmer errors with no modules to test."""
        with self.assertRaises(ValueError):
            LazyImportTester([])

    def test_subprocess_fails_with_no_command(self):
        """Catch programmer errors with no command to test."""
        with self.assertRaises(ValueError):
            LazySubprocessTester([])



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2018.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Pass manager for optimization level 0, providing no explicit optimization.

Level 0 pass manager: no explicit optimization other than mapping to backend.
"""

from qiskit.transpiler.passmanager_config import PassManagerConfig
from qiskit.transpiler.passmanager import StagedPassManager
from qiskit.transpiler.preset_passmanagers import common
from qiskit.transpiler.preset_passmanagers.plugin import (
    PassManagerStagePluginManager,
)


def level_0_pass_manager(pass_manager_config: PassManagerConfig) -> StagedPassManager:
    """Level 0 pass manager: no explicit optimization other than mapping to backend.

    This pass manager applies the user-given initial layout. If none is given, a trivial
    layout consisting of mapping the i-th virtual qubit to the i-th physical qubit is used.
    Any unused physical qubit is allocated as ancilla space.

    The pass manager then unrolls the circuit to the desired basis, and transforms the
    circuit to match the coupling map.

    Args:
        pass_manager_config: configuration of the pass manager.

    Returns:
        a level 0 pass manager.

    Raises:
        TranspilerError: if the passmanager config is invalid.
    """
    plugin_manager = PassManagerStagePluginManager()
    basis_gates = pass_manager_config.basis_gates
    coupling_map = pass_manager_config.coupling_map
    initial_layout = pass_manager_config.initial_layout
    init_method = pass_manager_config.init_method or "default"
    layout_method = pass_manager_config.layout_method or "default"
    routing_method = pass_manager_config.routing_method or "sabre"
    translation_method = pass_manager_config.translation_method or "translator"
    optimization_method = pass_manager_config.optimization_method or "default"
    scheduling_method = pass_manager_config.scheduling_method or "default"
    target = pass_manager_config.target

    # Choose routing pass
    routing_pm = plugin_manager.get_passmanager_stage(
        "routing", routing_method, pass_manager_config, optimization_level=0
    )

    # Build pass manager
    if coupling_map or initial_layout:
        layout = plugin_manager.get_passmanager_stage(
            "layout", layout_method, pass_manager_config, optimization_level=0
        )
        routing = routing_pm
    else:
        layout = None
        routing = None

    translation = plugin_manager.get_passmanager_stage(
        "translation", translation_method, pass_manager_config, optimization_level=0
    )

    if (coupling_map and not coupling_map.is_symmetric) or (
        target is not None and target.get_non_global_operation_names(strict_direction=True)
    ):
        pre_opt = common.generate_pre_op_passmanager(target, coupling_map)
        pre_opt += translation
    else:
        pre_opt = None

    sched = plugin_manager.get_passmanager_stage(
        "scheduling", scheduling_method, pass_manager_config, optimization_level=0
    )

    pre_init = common.generate_control_flow_options_check(
        layout_method=layout_method,
        routing_method=routing_method,
        translation_method=translation_method,
        optimization_method=optimization_method,
        scheduling_method=scheduling_method,
        basis_gates=basis_gates,
        target=target,
    )
    init = plugin_manager.get_passmanager_stage(
        "init", init_method, pass_manager_config, optimization_level=0
    )
    optimization = plugin_manager.get_passmanager_stage(
        "optimization", optimization_method, pass_manager_config, optimization_level=0
    )

    return StagedPassManager(
        pre_init=pre_init,
        init=init,
        layout=layout,
        routing=routing,
        translation=translation,
        pre_optimization=pre_opt,
        optimization=optimization,
        scheduling=sched,
    )



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test permutation synthesis functions."""


import unittest
import numpy as np
from ddt import ddt, data

from qiskit.quantum_info.operators import Operator
from qiskit.circuit.library import LinearFunction, PermutationGate
from qiskit.synthesis.permutation import (
    synth_permutation_acg,
    synth_permutation_depth_lnn_kms,
    synth_permutation_basic,
    synth_permutation_reverse_lnn_kms,
)
from qiskit.synthesis.permutation.permutation_utils import (
    _inverse_pattern,
    _validate_permutation,
)
from test import QiskitTestCase  # pylint: disable=wrong-import-order


@ddt
class TestPermutationSynthesis(QiskitTestCase):
    """Test the permutation synthesis functions."""

    @data(4, 5, 10, 15, 20)
    def test_inverse_pattern(self, width):
        """Test _inverse_pattern function produces correct index map."""
        np.random.seed(1)
        for _ in range(5):
            pattern = np.random.permutation(width)
            inverse = _inverse_pattern(pattern)
            for ii, jj in enumerate(pattern):
                self.assertTrue(inverse[jj] == ii)

    @data(10, 20)
    def test_invalid_permutations(self, width):
        """Check that _validate_permutation raises exceptions when the
        input is not a permutation."""
        np.random.seed(1)
        for _ in range(5):
            pattern = np.random.permutation(width)

            pattern_out_of_range = np.copy(pattern)
            pattern_out_of_range[0] = -1
            with self.assertRaises(ValueError) as exc:
                _validate_permutation(pattern_out_of_range)
                self.assertIn("input contains a negative number", str(exc.exception))

            pattern_out_of_range = np.copy(pattern)
            pattern_out_of_range[0] = width
            with self.assertRaises(ValueError) as exc:
                _validate_permutation(pattern_out_of_range)
                self.assertIn(f"input has length {width} and contains {width}", str(exc.exception))

            pattern_duplicate = np.copy(pattern)
            pattern_duplicate[-1] = pattern[0]
            with self.assertRaises(ValueError) as exc:
                _validate_permutation(pattern_duplicate)
                self.assertIn(f"input contains {pattern[0]} more than once", str(exc.exception))

    @data(4, 5, 10, 15, 20)
    def test_synth_permutation_basic(self, width):
        """Test synth_permutation_basic function produces the correct
        circuit."""
        np.random.seed(1)
        for _ in range(5):
            pattern = np.random.permutation(width)
            qc = synth_permutation_basic(pattern)

            # Check that the synthesized circuit consists of SWAP gates only.
            for instruction in qc.data:
                self.assertEqual(instruction.operation.name, "swap")

            # Construct a linear function from the synthesized circuit, and
            # check that its permutation pattern matches the original pattern.
            synthesized_pattern = LinearFunction(qc).permutation_pattern()
            self.assertTrue(np.array_equal(synthesized_pattern, pattern))

    @data(4, 5, 10, 15, 20)
    def test_synth_permutation_acg(self, width):
        """Test synth_permutation_acg function produces the correct
        circuit."""
        np.random.seed(1)
        for _ in range(5):
            pattern = np.random.permutation(width)
            qc = synth_permutation_acg(pattern)

            # Check that the synthesized circuit consists of SWAP gates only.
            for instruction in qc.data:
                self.assertEqual(instruction.operation.name, "swap")

            # Check that the depth of the circuit (measured in terms of SWAPs) is at most 2.
            self.assertLessEqual(qc.depth(), 2)

            # Construct a linear function from the synthesized circuit, and
            # check that its permutation pattern matches the original pattern.
            synthesized_pattern = LinearFunction(qc).permutation_pattern()
            self.assertTrue(np.array_equal(synthesized_pattern, pattern))

    @data(4, 5, 10, 15, 20)
    def test_synth_permutation_depth_lnn_kms(self, width):
        """Test synth_permutation_depth_lnn_kms function produces the correct
        circuit."""
        np.random.seed(1)
        for _ in range(5):
            pattern = np.random.permutation(width)
            qc = synth_permutation_depth_lnn_kms(pattern)

            # Check that the synthesized circuit consists of SWAP gates only,
            # and that these SWAPs adhere to the LNN connectivity.
            for instruction in qc.data:
                self.assertEqual(instruction.operation.name, "swap")
                q0 = qc.find_bit(instruction.qubits[0]).index
                q1 = qc.find_bit(instruction.qubits[1]).index
                dist = abs(q0 - q1)
                self.assertEqual(dist, 1)

            # Check that the depth of the circuit (measured in #SWAPs)
            # does not exceed the number of qubits.
            self.assertLessEqual(qc.depth(), width)

            # Construct a linear function from the synthesized circuit, and
            # check that its permutation pattern matches the original pattern.
            synthesized_pattern = LinearFunction(qc).permutation_pattern()
            self.assertTrue(np.array_equal(synthesized_pattern, pattern))

    @data(1, 2, 3, 4, 5, 10, 15, 20)
    def test_synth_permutation_reverse_lnn_kms(self, num_qubits):
        """Test synth_permutation_reverse_lnn_kms function produces the correct
        circuit."""
        pattern = list(reversed(range(num_qubits)))
        qc = synth_permutation_reverse_lnn_kms(num_qubits)
        self.assertListEqual((LinearFunction(qc).permutation_pattern()).tolist(), pattern)

        # Check that the CX depth of the circuit is at 2*n+2
        self.assertTrue(qc.depth() <= 2 * num_qubits + 2)

        # Check that the synthesized circuit consists of CX gates only,
        # and that these CXs adhere to the LNN connectivity.
        for instruction in qc.data:
            self.assertEqual(instruction.operation.name, "cx")
            q0 = qc.find_bit(instruction.qubits[0]).index
            q1 = qc.find_bit(instruction.qubits[1]).index
            dist = abs(q0 - q1)
            self.assertEqual(dist, 1)

    @data(4, 5, 6, 7)
    def test_permutation_matrix(self, width):
        """Test that the unitary matrix constructed from permutation pattern
        is correct."""
        np.random.seed(1)
        for _ in range(5):
            pattern = np.random.permutation(width)
            qc = synth_permutation_depth_lnn_kms(pattern)
            expected = Operator(qc)
            constructed = Operator(PermutationGate(pattern))
            self.assertEqual(expected, constructed)


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for drawing of timeline drawer."""

import numpy as np

import qiskit
from qiskit.visualization.timeline import drawings, types
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class TestDrawingObjects(QiskitTestCase):
    """Tests for drawings."""

    def setUp(self) -> None:
        """Setup."""
        super().setUp()

        # bits
        self.qubits = list(qiskit.QuantumRegister(2))

        # metadata
        self.meta1 = {"val1": 0, "val2": 1}
        self.meta2 = {"val1": 2, "val2": 3}

        # style data
        self.style1 = {"property1": 0, "property2": 1}
        self.style2 = {"property1": 2, "property2": 3}

    def test_line_data_equivalent(self):
        """Test LineData equivalent check."""
        xs = list(np.arange(10))
        ys = list(np.ones(10))

        obj1 = drawings.LineData(
            data_type=types.LineType.BARRIER,
            bit=self.qubits[0],
            xvals=xs,
            yvals=ys,
            meta=self.meta1,
            styles=self.style1,
        )

        obj2 = drawings.LineData(
            data_type=types.LineType.BARRIER,
            bit=self.qubits[0],
            xvals=xs,
            yvals=ys,
            meta=self.meta2,
            styles=self.style2,
        )

        self.assertEqual(obj1, obj2)

    def test_line_data_equivalent_abstract_coord(self):
        """Test LineData equivalent check with abstract coordinate."""
        xs = [types.AbstractCoordinate.LEFT, types.AbstractCoordinate.RIGHT]
        ys = [types.AbstractCoordinate.BOTTOM, types.AbstractCoordinate.TOP]

        obj1 = drawings.LineData(
            data_type=types.LineType.BARRIER,
            bit=self.qubits[0],
            xvals=xs,
            yvals=ys,
            meta=self.meta1,
            styles=self.style1,
        )

        obj2 = drawings.LineData(
            data_type=types.LineType.BARRIER,
            bit=self.qubits[0],
            xvals=xs,
            yvals=ys,
            meta=self.meta2,
            styles=self.style2,
        )

        self.assertEqual(obj1, obj2)

    def test_box_data_equivalent(self):
        """Test BoxData equivalent check."""
        xs = [0, 1]
        ys = [0, 1]

        obj1 = drawings.BoxData(
            data_type=types.BoxType.SCHED_GATE,
            bit=self.qubits[0],
            xvals=xs,
            yvals=ys,
            meta=self.meta1,
            styles=self.style1,
        )

        obj2 = drawings.BoxData(
            data_type=types.BoxType.SCHED_GATE,
            bit=self.qubits[0],
            xvals=xs,
            yvals=ys,
            meta=self.meta2,
            styles=self.style2,
        )

        self.assertEqual(obj1, obj2)

    def test_box_data_equivalent_abstract_coord(self):
        """Test BoxData equivalent check with abstract coordinate."""
        xs = [types.AbstractCoordinate.LEFT, types.AbstractCoordinate.RIGHT]
        ys = [types.AbstractCoordinate.BOTTOM, types.AbstractCoordinate.TOP]

        obj1 = drawings.BoxData(
            data_type=types.BoxType.SCHED_GATE,
            bit=self.qubits[0],
            xvals=xs,
            yvals=ys,
            meta=self.meta1,
            styles=self.style1,
        )

        obj2 = drawings.BoxData(
            data_type=types.BoxType.SCHED_GATE,
            bit=self.qubits[0],
            xvals=xs,
            yvals=ys,
            meta=self.meta2,
            styles=self.style2,
        )

        self.assertEqual(obj1, obj2)

    def test_text_data_equivalent(self):
        """Test TextData equivalent check."""
        obj1 = drawings.TextData(
            data_type=types.LabelType.GATE_NAME,
            bit=self.qubits[0],
            xval=0,
            yval=0,
            text="test",
            latex="test",
            meta=self.meta1,
            styles=self.style1,
        )

        obj2 = drawings.TextData(
            data_type=types.LabelType.GATE_NAME,
            bit=self.qubits[0],
            xval=0,
            yval=0,
            text="test",
            latex="test",
            meta=self.meta2,
            styles=self.style2,
        )

        self.assertEqual(obj1, obj2)

    def test_text_data_equivalent_abstract_coord(self):
        """Test TextData equivalent check with abstract coordinate."""
        obj1 = drawings.TextData(
            data_type=types.LabelType.GATE_NAME,
            bit=self.qubits[0],
            xval=types.AbstractCoordinate.LEFT,
            yval=types.AbstractCoordinate.BOTTOM,
            text="test",
            latex="test",
            meta=self.meta1,
            styles=self.style1,
        )

        obj2 = drawings.TextData(
            data_type=types.LabelType.GATE_NAME,
            bit=self.qubits[0],
            xval=types.AbstractCoordinate.LEFT,
            yval=types.AbstractCoordinate.BOTTOM,
            text="test",
            latex="test",
            meta=self.meta2,
            styles=self.style2,
        )

        self.assertEqual(obj1, obj2)

    def test_bit_link_data_equivalent(self):
        """Test BitLinkData equivalent check."""
        obj1 = drawings.GateLinkData(
            bits=[self.qubits[0], self.qubits[1]], xval=0, styles=self.style1
        )

        obj2 = drawings.GateLinkData(
            bits=[self.qubits[0], self.qubits[1]], xval=0, styles=self.style2
        )

        self.assertEqual(obj1, obj2)

    def test_bit_link_data_equivalent_abstract_coord(self):
        """Test BitLinkData equivalent check with abstract coordinate."""
        obj1 = drawings.GateLinkData(
            bits=[self.qubits[0], self.qubits[1]],
            xval=types.AbstractCoordinate.LEFT,
            styles=self.style1,
        )

        obj2 = drawings.GateLinkData(
            bits=[self.qubits[0], self.qubits[1]],
            xval=types.AbstractCoordinate.LEFT,
            styles=self.style2,
        )

        self.assertEqual(obj1, obj2)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2018.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Map a DAGCircuit onto a ``coupling_map`` adding swap gates."""

import itertools
import logging
from math import inf
import numpy as np

from qiskit.converters import dag_to_circuit, circuit_to_dag
from qiskit.circuit.classical import expr, types
from qiskit.circuit.quantumregister import QuantumRegister
from qiskit.transpiler.basepasses import TransformationPass
from qiskit.transpiler.exceptions import TranspilerError
from qiskit.dagcircuit import DAGCircuit
from qiskit.circuit.library.standard_gates import SwapGate
from qiskit.transpiler.layout import Layout
from qiskit.transpiler.target import Target
from qiskit.circuit import (
    Clbit,
    IfElseOp,
    WhileLoopOp,
    ForLoopOp,
    SwitchCaseOp,
    ControlFlowOp,
    CASE_DEFAULT,
)
from qiskit._accelerate import stochastic_swap as stochastic_swap_rs
from qiskit._accelerate import nlayout
from qiskit.transpiler.passes.layout import disjoint_utils
from qiskit.utils import deprecate_func

from .utils import get_swap_map_dag

logger = logging.getLogger(__name__)


class StochasticSwap(TransformationPass):
    """Map a DAGCircuit onto a `coupling_map` adding swap gates.

    Uses a randomized algorithm.

    Notes:
        1. Measurements may occur and be followed by swaps that result in repeated
           measurement of the same qubit. Near-term experiments cannot implement
           these circuits, so some care is required when using this mapper
           with experimental backend targets.

        2. We do not use the fact that the input state is zero to simplify
           the circuit.
    """

    @deprecate_func(
        since="1.3",
        removal_timeline="in the 2.0 release",
        additional_msg="The StochasticSwap transpilation pass is a suboptimal "
        "routing algorithm and has been superseded by the SabreSwap pass.",
    )
    def __init__(self, coupling_map, trials=20, seed=None, fake_run=False, initial_layout=None):
        """StochasticSwap initializer.

        The coupling map is a connected graph

        If these are not satisfied, the behavior is undefined.

        Args:
            coupling_map (Union[CouplingMap, Target]): Directed graph representing a coupling
                map.
            trials (int): maximum number of iterations to attempt
            seed (int): seed for random number generator
            fake_run (bool): if true, it will only pretend to do routing, i.e., no
                swap is effectively added.
            initial_layout (Layout): starting layout at beginning of pass.
        """
        super().__init__()

        if isinstance(coupling_map, Target):
            self.target = coupling_map
            self.coupling_map = self.target.build_coupling_map()
        else:
            self.target = None
            self.coupling_map = coupling_map
        self.trials = trials
        self.seed = seed
        self.rng = None
        self.fake_run = fake_run
        self.qregs = None
        self.initial_layout = initial_layout
        self._int_to_qubit = None

    def run(self, dag):
        """Run the StochasticSwap pass on `dag`.

        Args:
            dag (DAGCircuit): DAG to map.

        Returns:
            DAGCircuit: A mapped DAG.

        Raises:
            TranspilerError: if the coupling map or the layout are not
            compatible with the DAG, or if the coupling_map=None
        """

        if self.coupling_map is None:
            raise TranspilerError("StochasticSwap cannot run with coupling_map=None")

        if len(dag.qregs) != 1 or dag.qregs.get("q", None) is None:
            raise TranspilerError("StochasticSwap runs on physical circuits only")

        if len(dag.qubits) > len(self.coupling_map.physical_qubits):
            raise TranspilerError("The layout does not match the amount of qubits in the DAG")
        disjoint_utils.require_layout_isolated_to_component(
            dag, self.coupling_map if self.target is None else self.target
        )

        self.rng = np.random.default_rng(self.seed)

        canonical_register = dag.qregs["q"]
        if self.initial_layout is None:
            self.initial_layout = Layout.generate_trivial_layout(canonical_register)
        # Qubit indices are used to assign an integer to each virtual qubit during the routing: it's
        # a mapping of {virtual: virtual}, for converting between Python and Rust forms.
        self._int_to_qubit = tuple(dag.qubits)

        self.qregs = dag.qregs
        logger.debug("StochasticSwap rng seeded with seed=%s", self.seed)
        self.coupling_map.compute_distance_matrix()
        new_dag = self._mapper(dag, self.coupling_map, trials=self.trials)
        return new_dag

    def _layer_permutation(self, dag, layer_partition, layout, qubit_subset, coupling, trials):
        """Find a swap circuit that implements a permutation for this layer.

        The goal is to swap qubits such that qubits in the same two-qubit gates
        are adjacent.

        Based on S. Bravyi's algorithm.

        Args:
            layer_partition (list): The layer_partition is a list of (qu)bit
                lists and each qubit is a tuple (qreg, index).
            layout (Layout): The layout is a Layout object mapping virtual
                qubits in the input circuit to physical qubits in the coupling
                graph. It reflects the current positions of the data.
            qubit_subset (list): The qubit_subset is the set of qubits in
                the coupling graph that we have chosen to map into, as tuples
                (Register, index).
            coupling (CouplingMap): Directed graph representing a coupling map.
                This coupling map should be one that was provided to the
                stochastic mapper.
            trials (int): Number of attempts the randomized algorithm makes.

        Returns:
            Tuple: success_flag, best_circuit, best_depth, best_layout

        If success_flag is True, then best_circuit contains a DAGCircuit with
        the swap circuit, best_depth contains the depth of the swap circuit,
        and best_layout contains the new positions of the data qubits after the
        swap circuit has been applied.

        Raises:
            TranspilerError: if anything went wrong.
        """
        logger.debug("layer_permutation: layer_partition = %s", layer_partition)
        logger.debug("layer_permutation: layout = %s", layout.get_virtual_bits())
        logger.debug("layer_permutation: qubit_subset = %s", qubit_subset)
        logger.debug("layer_permutation: trials = %s", trials)

        # The input dag is on a flat canonical register
        canonical_register = QuantumRegister(len(layout), "q")

        gates = []  # list of lists of tuples [[(register, index), ...], ...]
        for gate_args in layer_partition:
            if len(gate_args) > 2:
                raise TranspilerError("Layer contains > 2-qubit gates")
            if len(gate_args) == 2:
                gates.append(tuple(gate_args))
        logger.debug("layer_permutation: gates = %s", gates)

        # Can we already apply the gates? If so, there is no work to do.
        # Accessing via private attributes to avoid overhead from __getitem__
        # and to optimize performance of the distance matrix access
        dist = sum(coupling._dist_matrix[layout._v2p[g[0]], layout._v2p[g[1]]] for g in gates)
        logger.debug("layer_permutation: distance = %s", dist)
        if dist == len(gates):
            logger.debug("layer_permutation: nothing to do")
            circ = DAGCircuit()
            circ.add_qreg(canonical_register)
            return True, circ, 0, layout

        # Begin loop over trials of randomized algorithm
        num_qubits = len(layout)
        best_depth = inf  # initialize best depth
        best_edges = None  # best edges found
        best_circuit = None  # initialize best swap circuit
        best_layout = None  # initialize best final layout

        cdist2 = coupling._dist_matrix**2
        int_qubit_subset = np.fromiter(
            (dag.find_bit(bit).index for bit in qubit_subset),
            dtype=np.uint32,
            count=len(qubit_subset),
        )

        int_gates = np.fromiter(
            (dag.find_bit(bit).index for gate in gates for bit in gate),
            dtype=np.uint32,
            count=2 * len(gates),
        )

        layout_mapping = {dag.find_bit(k).index: v for k, v in layout.get_virtual_bits().items()}
        int_layout = nlayout.NLayout(layout_mapping, num_qubits, coupling.size())

        trial_circuit = DAGCircuit()  # SWAP circuit for slice of swaps in this trial
        trial_circuit.add_qubits(list(layout.get_virtual_bits()))

        edges = np.asarray(coupling.get_edges(), dtype=np.uint32).ravel()
        cdist = coupling._dist_matrix
        best_edges, best_layout, best_depth = stochastic_swap_rs.swap_trials(
            trials,
            num_qubits,
            int_layout,
            int_qubit_subset,
            int_gates,
            cdist,
            cdist2,
            edges,
            seed=self.seed,
        )
        # If we have no best circuit for this layer, all of the trials have failed
        if best_layout is None:
            logger.debug("layer_permutation: failed!")
            return False, None, None, None

        edges = best_edges.edges()
        for idx in range(len(edges) // 2):
            swap_src = self._int_to_qubit[edges[2 * idx]]
            swap_tgt = self._int_to_qubit[edges[2 * idx + 1]]
            trial_circuit.apply_operation_back(SwapGate(), (swap_src, swap_tgt), (), check=False)
        best_circuit = trial_circuit

        # Otherwise, we return our result for this layer
        logger.debug("layer_permutation: success!")
        layout_mapping = best_layout.layout_mapping()

        best_lay = Layout({best_circuit.qubits[k]: v for (k, v) in layout_mapping})
        return True, best_circuit, best_depth, best_lay

    def _layer_update(self, dag, layer, best_layout, best_depth, best_circuit):
        """Add swaps followed by the now mapped layer from the original circuit.

        Args:
            dag (DAGCircuit): The DAGCircuit object that the _mapper method is building
            layer (DAGCircuit): A DAGCircuit layer from the original circuit
            best_layout (Layout): layout returned from _layer_permutation
            best_depth (int): depth returned from _layer_permutation
            best_circuit (DAGCircuit): swap circuit returned from _layer_permutation
        """
        logger.debug("layer_update: layout = %s", best_layout)
        logger.debug("layer_update: self.initial_layout = %s", self.initial_layout)

        # Output any swaps
        if best_depth > 0:
            logger.debug("layer_update: there are swaps in this layer, depth %d", best_depth)
            dag.compose(best_circuit, qubits=list(best_circuit.qubits), inline_captures=True)
        else:
            logger.debug("layer_update: there are no swaps in this layer")
        # Output this layer
        dag.compose(
            layer["graph"], qubits=best_layout.reorder_bits(dag.qubits), inline_captures=True
        )

    def _mapper(self, circuit_graph, coupling_graph, trials=20):
        """Map a DAGCircuit onto a CouplingMap using swap gates.

        Args:
            circuit_graph (DAGCircuit): input DAG circuit
            coupling_graph (CouplingMap): coupling graph to map onto
            trials (int): number of trials.

        Returns:
            DAGCircuit: object containing a circuit equivalent to
                circuit_graph that respects couplings in coupling_graph

        Raises:
            TranspilerError: if there was any error during the mapping
                or with the parameters.
        """
        # Schedule the input circuit by calling layers()
        layerlist = list(circuit_graph.layers())
        logger.debug("schedule:")
        for i, v in enumerate(layerlist):
            logger.debug("    %d: %s", i, v["partition"])

        qubit_subset = self.initial_layout.get_virtual_bits().keys()

        # Find swap circuit to precede each layer of input circuit
        layout = self.initial_layout.copy()

        # Construct an empty DAGCircuit with the same set of
        # qregs and cregs as the input circuit
        dagcircuit_output = None
        if not self.fake_run:
            dagcircuit_output = circuit_graph.copy_empty_like()

        logger.debug("layout = %s", layout)

        # Iterate over layers
        for i, layer in enumerate(layerlist):
            # First try and compute a route for the entire layer in one go.
            if not layer["graph"].op_nodes(op=ControlFlowOp):
                success_flag, best_circuit, best_depth, best_layout = self._layer_permutation(
                    circuit_graph, layer["partition"], layout, qubit_subset, coupling_graph, trials
                )

                logger.debug("mapper: layer %d", i)
                logger.debug("mapper: success_flag=%s,best_depth=%s", success_flag, str(best_depth))
                if success_flag:
                    layout = best_layout

                    # Update the DAG
                    if not self.fake_run:
                        self._layer_update(
                            dagcircuit_output, layer, best_layout, best_depth, best_circuit
                        )
                    continue

            # If we're here, we need to go through every gate in the layer serially.
            logger.debug("mapper: failed, layer %d, retrying sequentially", i)
            # Go through each gate in the layer
            for j, serial_layer in enumerate(layer["graph"].serial_layers()):
                layer_dag = serial_layer["graph"]
                # layer_dag has only one operation
                op_node = layer_dag.op_nodes()[0]
                if isinstance(op_node.op, ControlFlowOp):
                    layout = self._controlflow_layer_update(
                        dagcircuit_output, layer_dag, layout, circuit_graph
                    )
                else:
                    (success_flag, best_circuit, best_depth, best_layout) = self._layer_permutation(
                        circuit_graph,
                        serial_layer["partition"],
                        layout,
                        qubit_subset,
                        coupling_graph,
                        trials,
                    )
                    logger.debug("mapper: layer %d, sublayer %d", i, j)
                    logger.debug(
                        "mapper: success_flag=%s,best_depth=%s,", success_flag, str(best_depth)
                    )

                    # Give up if we fail again
                    if not success_flag:
                        raise TranspilerError(f"swap mapper failed: layer {i}, sublayer {j}")

                    # Update the record of qubit positions
                    # for each inner iteration
                    layout = best_layout
                    # Update the DAG
                    if not self.fake_run:
                        self._layer_update(
                            dagcircuit_output,
                            serial_layer,
                            best_layout,
                            best_depth,
                            best_circuit,
                        )

        # This is the final edgemap. We might use it to correctly replace
        # any measurements that needed to be removed earlier.
        logger.debug("mapper: self.initial_layout = %s", self.initial_layout)
        logger.debug("mapper: layout = %s", layout)
        if self.property_set["final_layout"] is None:
            self.property_set["final_layout"] = layout
        else:
            self.property_set["final_layout"] = layout.compose(
                self.property_set["final_layout"], circuit_graph.qubits
            )

        if self.fake_run:
            return circuit_graph
        return dagcircuit_output

    def _controlflow_layer_update(self, dagcircuit_output, layer_dag, current_layout, root_dag):
        """
        Updates the new dagcircuit with a routed control flow operation.

        Args:
           dagcircuit_output (DAGCircuit): dagcircuit that is being built with routed operations.
           layer_dag (DAGCircuit): layer to route containing a single controlflow operation.
           current_layout (Layout): current layout coming into this layer.
           root_dag (DAGCircuit): root dag of pass

        Returns:
           Layout: updated layout after this layer has been routed.

        Raises:
            TranspilerError: if layer_dag does not contain a recognized ControlFlowOp.

        """
        node = layer_dag.op_nodes()[0]
        if not isinstance(node.op, (IfElseOp, ForLoopOp, WhileLoopOp, SwitchCaseOp)):
            raise TranspilerError(f"unsupported control flow operation: {node}")
        # For each block, expand it up be the full width of the containing DAG so we can be certain
        # that it is routable, then route it within that.  When we recombine later, we'll reduce all
        # these blocks down to remove any qubits that are idle.
        block_dags = []
        block_layouts = []
        for block in node.op.blocks:
            inner_pass = self._recursive_pass(current_layout)
            block_dags.append(inner_pass.run(_dag_from_block(block, node, root_dag)))
            block_layouts.append(inner_pass.property_set["final_layout"].copy())

        # Determine what layout we need to go towards.  For some blocks (such as `for`), we must
        # guarantee that the final layout is the same as the initial or the loop won't work.
        if _controlflow_exhaustive_acyclic(node.op):
            # We heuristically just choose to use the layout of whatever the deepest block is, to
            # avoid extending the total depth by too much.
            final_layout = max(
                zip(block_layouts, block_dags), key=lambda x: x[1].depth(recurse=True)
            )[0]
        else:
            final_layout = current_layout
        if self.fake_run:
            return final_layout

        # Add swaps to the end of each block to make sure they all have the same layout at the end.
        # Adding these swaps can cause fewer wires to be idle than we expect (if we have to swap
        # across unused qubits), so we track that at this point too.
        idle_qubits = set(root_dag.qubits)
        for layout, updated_dag_block in zip(block_layouts, block_dags):
            swap_dag, swap_qubits = get_swap_map_dag(
                root_dag, self.coupling_map, layout, final_layout, seed=self._new_seed()
            )
            if swap_dag.size(recurse=False):
                updated_dag_block.compose(swap_dag, qubits=swap_qubits, inline_captures=True)
            idle_qubits &= set(updated_dag_block.idle_wires())

        # Now for each block, expand it to be full width over all active wires (all blocks of a
        # control-flow operation need to have equal input wires), and convert it to circuit form.
        block_circuits = []
        for updated_dag_block in block_dags:
            updated_dag_block.remove_qubits(*idle_qubits)
            block_circuits.append(dag_to_circuit(updated_dag_block))

        new_op = node.op.replace_blocks(block_circuits)
        new_qargs = block_circuits[0].qubits
        dagcircuit_output.apply_operation_back(new_op, new_qargs, node.cargs, check=False)
        return final_layout

    def _new_seed(self):
        """Get a seed for a new RNG instance."""
        return self.rng.integers(0x7FFF_FFFF_FFFF_FFFF)

    def _recursive_pass(self, initial_layout):
        """Get a new instance of this class to handle a recursive call for a control-flow block.

        Each pass starts with its own new seed, determined deterministically from our own."""
        return self.__class__(
            self.coupling_map,
            # This doesn't cause an exponential explosion of the trials because we only generate a
            # recursive pass instance for control-flow operations, while the trial multiplicity is
            # only for non-control-flow layers.
            trials=self.trials,
            seed=self._new_seed(),
            fake_run=self.fake_run,
            initial_layout=initial_layout,
        )


def _controlflow_exhaustive_acyclic(operation: ControlFlowOp):
    """Return True if the entire control-flow operation represents a block that is guaranteed to be
    entered, and does not cycle back to the initial layout."""
    if isinstance(operation, IfElseOp):
        return len(operation.blocks) == 2
    if isinstance(operation, SwitchCaseOp):
        cases = operation.cases()
        if isinstance(operation.target, expr.Expr):
            type_ = operation.target.type
            if type_.kind is types.Bool:
                max_matches = 2
            elif type_.kind is types.Uint:
                max_matches = 1 << type_.width
            else:
                raise RuntimeError(f"unhandled target type: '{type_}'")
        else:
            max_matches = 2 if isinstance(operation.target, Clbit) else 1 << len(operation.target)
        return CASE_DEFAULT in cases or len(cases) == max_matches
    return False


def _dag_from_block(block, node, root_dag):
    """Get a :class:`DAGCircuit` that represents the :class:`.QuantumCircuit` ``block`` embedded
    within the ``root_dag`` for full-width routing purposes.  This means that all the qubits are in
    the output DAG, but only the necessary clbits and classical registers are."""
    out = DAGCircuit()
    # The pass already ensured that `root_dag` has only a single quantum register with everything.
    for qreg in root_dag.qregs.values():
        out.add_qreg(qreg)
    # For clbits, we need to take more care.  Nested control-flow might need registers to exist for
    # conditions on inner blocks.  `DAGCircuit.substitute_node_with_dag` handles this register
    # mapping when required, so we use that with a dummy block that pretends to act on all variables
    # in the DAG.
    out.add_clbits(node.cargs)
    for var in block.iter_input_vars():
        out.add_input_var(var)
    for var in block.iter_captured_vars():
        out.add_captured_var(var)
    for var in block.iter_declared_vars():
        out.add_declared_var(var)

    dummy = out.apply_operation_back(
        IfElseOp(expr.lift(True), block.copy_empty_like(vars_mode="captures")),
        node.qargs,
        node.cargs,
        check=False,
    )
    wire_map = dict(itertools.chain(zip(block.qubits, node.qargs), zip(block.clbits, node.cargs)))
    out.substitute_node_with_dag(dummy, circuit_to_dag(block), wires=wire_map)
    return out



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Unit tests for pulse waveforms."""

import unittest
from unittest.mock import patch
import numpy as np
import symengine as sym

from qiskit.circuit import Parameter
from qiskit.pulse.library import (
    SymbolicPulse,
    ScalableSymbolicPulse,
    Waveform,
    Constant,
    Gaussian,
    GaussianSquare,
    GaussianSquareDrag,
    gaussian_square_echo,
    GaussianDeriv,
    Drag,
    Sin,
    Cos,
    Sawtooth,
    Triangle,
    Square,
    Sech,
    SechDeriv,
)
from qiskit.pulse import functional_pulse, PulseError
from test import QiskitTestCase  # pylint: disable=wrong-import-order
from qiskit.utils.deprecate_pulse import decorate_test_methods, ignore_pulse_deprecation_warnings


@decorate_test_methods(ignore_pulse_deprecation_warnings)
class TestWaveform(QiskitTestCase):
    """Waveform tests."""

    def test_sample_pulse(self):
        """Test pulse initialization."""
        n_samples = 100
        samples = np.linspace(0, 1.0, n_samples, dtype=np.complex128)
        name = "test"
        sample_pulse = Waveform(samples, name=name)

        self.assertEqual(sample_pulse.samples.dtype, np.complex128)
        np.testing.assert_almost_equal(sample_pulse.samples, samples)

        self.assertEqual(sample_pulse.duration, n_samples)
        self.assertEqual(sample_pulse.name, name)

    def test_waveform_hashing(self):
        """Test waveform hashing."""
        n_samples = 100
        samples = np.linspace(0, 1.0, n_samples, dtype=np.complex128)
        name = "test"
        sample_pulse = Waveform(samples, name=name)
        sample_pulse2 = Waveform(samples, name="test2")

        self.assertEqual({sample_pulse, sample_pulse2}, {sample_pulse})

    def test_type_casting(self):
        """Test casting of input samples to numpy array."""
        n_samples = 100
        samples_f64 = np.linspace(0, 1.0, n_samples, dtype=np.float64)

        sample_pulse_f64 = Waveform(samples_f64)
        self.assertEqual(sample_pulse_f64.samples.dtype, np.complex128)

        samples_c64 = np.linspace(0, 1.0, n_samples, dtype=np.complex64)

        sample_pulse_c64 = Waveform(samples_c64)
        self.assertEqual(sample_pulse_c64.samples.dtype, np.complex128)

        samples_list = np.linspace(0, 1.0, n_samples).tolist()

        sample_pulse_list = Waveform(samples_list)
        self.assertEqual(sample_pulse_list.samples.dtype, np.complex128)

    def test_pulse_limits(self):
        """Test that limits of pulse norm of one are enforced properly."""

        # test norm is correct for complex128 numpy data
        unit_pulse_c128 = np.exp(1j * 2 * np.pi * np.linspace(0, 1, 1000), dtype=np.complex128)
        # test does not raise error
        try:
            Waveform(unit_pulse_c128)
        except PulseError:
            self.fail("Waveform incorrectly failed on approximately unit norm samples.")

        invalid_const = 1.1
        with self.assertRaises(PulseError):
            Waveform(invalid_const * np.exp(1j * 2 * np.pi * np.linspace(0, 1, 1000)))

        with patch("qiskit.pulse.library.pulse.Pulse.limit_amplitude", new=False):
            wave = Waveform(invalid_const * np.exp(1j * 2 * np.pi * np.linspace(0, 1, 1000)))
            self.assertGreater(np.max(np.abs(wave.samples)), 1.0)

        # Test case where data is converted to python types with complex as a list
        # with form [re, im] and back to a numpy array.
        # This is how the transport layer handles samples in the qobj so it is important
        # to test.
        unit_pulse_c64 = np.exp(1j * 2 * np.pi * np.linspace(0, 1, 1000), dtype=np.complex64)
        sample_components = np.stack(
            np.transpose([np.real(unit_pulse_c64), np.imag(unit_pulse_c64)])
        )
        pulse_list = sample_components.tolist()
        recombined_pulse = [sample[0] + sample[1] * 1j for sample in pulse_list]

        # test does not raise error
        try:
            Waveform(recombined_pulse)
        except PulseError:
            self.fail("Waveform incorrectly failed to approximately unit norm samples.")


@decorate_test_methods(ignore_pulse_deprecation_warnings)
class TestSymbolicPulses(QiskitTestCase):
    """Tests for all subclasses of SymbolicPulse."""

    def test_construction(self):
        """Test that symbolic pulses can be constructed without error."""
        Gaussian(duration=25, sigma=4, amp=0.5, angle=np.pi / 2)
        GaussianSquare(duration=150, amp=0.2, sigma=8, width=140)
        GaussianSquare(duration=150, amp=0.2, sigma=8, risefall_sigma_ratio=2.5)
        Constant(duration=150, amp=0.5, angle=np.pi * 0.23)
        Drag(duration=25, amp=0.6, sigma=7.8, beta=4, angle=np.pi * 0.54)
        GaussianDeriv(duration=150, amp=0.2, sigma=8)
        Sin(duration=25, amp=0.5, freq=0.1, phase=0.5, angle=0.5)
        Cos(duration=30, amp=0.5, freq=0.1, phase=-0.5)
        Sawtooth(duration=40, amp=0.5, freq=0.2, phase=3.14)
        Triangle(duration=50, amp=0.5, freq=0.01, phase=0.5)
        Square(duration=50, amp=0.5, freq=0.01, phase=0.5)
        Sech(duration=50, amp=0.5, sigma=10)
        Sech(duration=50, amp=0.5, sigma=10, zero_ends=False)
        SechDeriv(duration=50, amp=0.5, sigma=10)

    def test_gauss_square_extremes(self):
        """Test that the gaussian square pulse can build a gaussian."""
        duration = 125
        sigma = 4
        amp = 0.5
        angle = np.pi / 2
        gaus_square = GaussianSquare(duration=duration, sigma=sigma, amp=amp, width=0, angle=angle)
        gaus = Gaussian(duration=duration, sigma=sigma, amp=amp, angle=angle)
        np.testing.assert_almost_equal(
            gaus_square.get_waveform().samples, gaus.get_waveform().samples
        )
        gaus_square = GaussianSquare(
            duration=duration, sigma=sigma, amp=amp, width=121, angle=angle
        )
        const = Constant(duration=duration, amp=amp, angle=angle)
        np.testing.assert_almost_equal(
            gaus_square.get_waveform().samples[2:-2], const.get_waveform().samples[2:-2]
        )

    def test_gauss_square_passes_validation_after_construction(self):
        """Test that parameter validation is consistent before and after construction.

        This previously used to raise an exception: see gh-7882."""
        pulse = GaussianSquare(duration=125, sigma=4, amp=0.5, width=100, angle=np.pi / 2)
        pulse.validate_parameters()

    def test_gaussian_square_drag_pulse(self):
        """Test that GaussianSquareDrag sample pulse matches expectations.

        Test that the real part of the envelop matches GaussianSquare and that
        the rise and fall match Drag.
        """
        risefall = 32
        sigma = 4
        amp = 0.5
        width = 100
        beta = 1
        duration = width + 2 * risefall

        gsd = GaussianSquareDrag(duration=duration, sigma=sigma, amp=amp, width=width, beta=beta)
        gsd_samples = gsd.get_waveform().samples

        gs_pulse = GaussianSquare(duration=duration, sigma=sigma, amp=amp, width=width)
        np.testing.assert_almost_equal(
            np.real(gsd_samples),
            np.real(gs_pulse.get_waveform().samples),
        )
        gsd2 = GaussianSquareDrag(
            duration=duration,
            sigma=sigma,
            amp=amp,
            beta=beta,
            risefall_sigma_ratio=risefall / sigma,
        )
        np.testing.assert_almost_equal(
            gsd_samples,
            gsd2.get_waveform().samples,
        )

        drag_pulse = Drag(duration=2 * risefall, amp=amp, sigma=sigma, beta=beta)
        np.testing.assert_almost_equal(
            gsd_samples[:risefall],
            drag_pulse.get_waveform().samples[:risefall],
        )
        np.testing.assert_almost_equal(
            gsd_samples[-risefall:],
            drag_pulse.get_waveform().samples[-risefall:],
        )

    def test_gauss_square_drag_extreme(self):
        """Test that the gaussian square drag pulse can build a drag pulse."""
        duration = 125
        sigma = 4
        amp = 0.5
        angle = 1.5
        beta = 1
        gsd = GaussianSquareDrag(
            duration=duration, sigma=sigma, amp=amp, width=0, beta=beta, angle=angle
        )
        drag = Drag(duration=duration, sigma=sigma, amp=amp, beta=beta, angle=angle)
        np.testing.assert_almost_equal(gsd.get_waveform().samples, drag.get_waveform().samples)

    def test_gaussian_square_drag_validation(self):
        """Test drag beta parameter validation."""

        GaussianSquareDrag(duration=50, width=0, sigma=16, amp=1, beta=2)
        GaussianSquareDrag(duration=50, width=0, sigma=16, amp=1, beta=4)
        GaussianSquareDrag(duration=50, width=0, sigma=16, amp=0.5, beta=20)
        GaussianSquareDrag(duration=50, width=0, sigma=16, amp=-1, beta=2)
        GaussianSquareDrag(duration=50, width=0, sigma=16, amp=1, beta=-2)
        GaussianSquareDrag(duration=50, width=0, sigma=16, amp=1, beta=6)
        GaussianSquareDrag(duration=50, width=0, sigma=16, amp=-0.5, beta=25, angle=1.5)
        with self.assertRaises(PulseError):
            GaussianSquareDrag(duration=50, width=0, sigma=16, amp=1, beta=20)
        with self.assertRaises(PulseError):
            GaussianSquareDrag(duration=50, width=0, sigma=4, amp=0.8, beta=20)
        with self.assertRaises(PulseError):
            GaussianSquareDrag(duration=50, width=0, sigma=4, amp=0.8, beta=-20)

    def test_gaussian_square_echo_pulse(self):
        """Test that gaussian_square_echo sample pulse matches expectations.

        Test that the real part of the envelop matches GaussianSquare with
        given amplitude and phase active for half duration with another
        GaussianSquare active for the other half duration with opposite
        amplitude and a GaussianSquare active on the entire duration with
        its own amplitude and phase
        """
        risefall = 32
        sigma = 4
        amp = 0.5
        width = 100
        duration = width + 2 * risefall
        active_amp = 0.1
        width_echo = (duration - 2 * (duration - width)) / 2

        gse = gaussian_square_echo(
            duration=duration, sigma=sigma, amp=amp, width=width, active_amp=active_amp
        )
        gse_samples = gse.get_waveform().samples

        gs_echo_pulse_pos = GaussianSquare(
            duration=duration / 2, sigma=sigma, amp=amp, width=width_echo
        )
        gs_echo_pulse_neg = GaussianSquare(
            duration=duration / 2, sigma=sigma, amp=-amp, width=width_echo
        )
        gs_active_pulse = GaussianSquare(
            duration=duration, sigma=sigma, amp=active_amp, width=width
        )
        gs_echo_pulse_pos_samples = np.array(
            gs_echo_pulse_pos.get_waveform().samples.tolist() + [0] * int(duration / 2)
        )
        gs_echo_pulse_neg_samples = np.array(
            [0] * int(duration / 2) + gs_echo_pulse_neg.get_waveform().samples.tolist()
        )
        gs_active_pulse_samples = gs_active_pulse.get_waveform().samples

        np.testing.assert_almost_equal(
            gse_samples,
            gs_echo_pulse_pos_samples + gs_echo_pulse_neg_samples + gs_active_pulse_samples,
        )

    def test_gaussian_square_echo_active_amp_validation(self):
        """Test gaussian square echo active amp parameter validation."""

        gaussian_square_echo(duration=50, width=0, sigma=16, amp=0.1, active_amp=0.2)
        gaussian_square_echo(duration=50, width=0, sigma=16, amp=0.1, active_amp=0.4)
        gaussian_square_echo(duration=50, width=0, sigma=16, amp=0.5, active_amp=0.3)
        gaussian_square_echo(duration=50, width=0, sigma=16, amp=-0.1, active_amp=0.2)
        gaussian_square_echo(duration=50, width=0, sigma=16, amp=0.1, active_amp=-0.2)
        gaussian_square_echo(duration=50, width=0, sigma=16, amp=0.1, active_amp=0.6)
        gaussian_square_echo(duration=50, width=0, sigma=16, amp=-0.5, angle=1.5, active_amp=0.25)
        with self.assertRaises(PulseError):
            gaussian_square_echo(duration=50, width=0, sigma=16, amp=0.1, active_amp=1.1)
        with self.assertRaises(PulseError):
            gaussian_square_echo(duration=50, width=0, sigma=4, amp=-0.8, active_amp=-0.3)

    def test_drag_validation(self):
        """Test drag parameter validation, specifically the beta validation."""
        duration = 25
        sigma = 4
        amp = 0.5
        angle = np.pi / 2
        beta = 1
        wf = Drag(duration=duration, sigma=sigma, amp=amp, beta=beta, angle=angle)
        samples = wf.get_waveform().samples
        self.assertTrue(max(np.abs(samples)) <= 1)
        with self.assertRaises(PulseError):
            wf = Drag(duration=duration, sigma=sigma, amp=1.2, beta=beta)
        beta = sigma**2
        with self.assertRaises(PulseError):
            wf = Drag(duration=duration, sigma=sigma, amp=amp, beta=beta, angle=angle)
        # If sigma is high enough, side peaks fall out of range and norm restriction is met
        sigma = 100
        wf = Drag(duration=duration, sigma=sigma, amp=amp, beta=beta, angle=angle)

    def test_drag_beta_validation(self):
        """Test drag beta parameter validation."""

        def check_drag(duration, sigma, amp, beta, angle=0):
            wf = Drag(duration=duration, sigma=sigma, amp=amp, beta=beta, angle=angle)
            samples = wf.get_waveform().samples
            self.assertTrue(max(np.abs(samples)) <= 1)

        check_drag(duration=50, sigma=16, amp=1, beta=2)
        check_drag(duration=50, sigma=16, amp=1, beta=4)
        check_drag(duration=50, sigma=16, amp=0.5, beta=20)
        check_drag(duration=50, sigma=16, amp=-1, beta=2)
        check_drag(duration=50, sigma=16, amp=1, beta=-2)
        check_drag(duration=50, sigma=16, amp=1, beta=6)
        check_drag(duration=50, sigma=16, amp=0.5, beta=25, angle=-np.pi / 2)
        with self.assertRaises(PulseError):
            check_drag(duration=50, sigma=16, amp=1, beta=20)
        with self.assertRaises(PulseError):
            check_drag(duration=50, sigma=4, amp=0.8, beta=20)
        with self.assertRaises(PulseError):
            check_drag(duration=50, sigma=4, amp=0.8, beta=-20)

    def test_sin_pulse(self):
        """Test that Sin creation"""
        duration = 100
        amp = 0.5
        freq = 0.1
        phase = 0

        Sin(duration=duration, amp=amp, freq=freq, phase=phase)

        with self.assertRaises(PulseError):
            Sin(duration=duration, amp=amp, freq=5, phase=phase)

    def test_cos_pulse(self):
        """Test that Cos creation"""
        duration = 100
        amp = 0.5
        freq = 0.1
        phase = 0
        cos_pulse = Cos(duration=duration, amp=amp, freq=freq, phase=phase)

        shifted_sin_pulse = Sin(duration=duration, amp=amp, freq=freq, phase=phase + np.pi / 2)
        np.testing.assert_almost_equal(
            shifted_sin_pulse.get_waveform().samples, cos_pulse.get_waveform().samples
        )
        with self.assertRaises(PulseError):
            Cos(duration=duration, amp=amp, freq=5, phase=phase)

    def test_square_pulse(self):
        """Test that Square pulse creation"""
        duration = 100
        amp = 0.5
        freq = 0.1
        phase = 0.3
        Square(duration=duration, amp=amp, freq=freq, phase=phase)

        with self.assertRaises(PulseError):
            Square(duration=duration, amp=amp, freq=5, phase=phase)

    def test_sawtooth_pulse(self):
        """Test that Sawtooth pulse creation"""
        duration = 100
        amp = 0.5
        freq = 0.1
        phase = 0.5
        sawtooth_pulse = Sawtooth(duration=duration, amp=amp, freq=freq, phase=phase)

        sawtooth_pulse_2 = Sawtooth(duration=duration, amp=amp, freq=freq, phase=phase + 2 * np.pi)
        np.testing.assert_almost_equal(
            sawtooth_pulse.get_waveform().samples, sawtooth_pulse_2.get_waveform().samples
        )

        with self.assertRaises(PulseError):
            Sawtooth(duration=duration, amp=amp, freq=5, phase=phase)

    def test_triangle_pulse(self):
        """Test that Triangle pulse creation"""
        duration = 100
        amp = 0.5
        freq = 0.1
        phase = 0.5
        triangle_pulse = Triangle(duration=duration, amp=amp, freq=freq, phase=phase)

        triangle_pulse_2 = Triangle(duration=duration, amp=amp, freq=freq, phase=phase + 2 * np.pi)
        np.testing.assert_almost_equal(
            triangle_pulse.get_waveform().samples, triangle_pulse_2.get_waveform().samples
        )

        with self.assertRaises(PulseError):
            Triangle(duration=duration, amp=amp, freq=5, phase=phase)

    def test_gaussian_deriv_pulse(self):
        """Test that GaussianDeriv pulse creation"""
        duration = 300
        amp = 0.5
        sigma = 100
        GaussianDeriv(duration=duration, amp=amp, sigma=sigma)

        with self.assertRaises(PulseError):
            Sech(duration=duration, amp=amp, sigma=0)

    def test_sech_pulse(self):
        """Test that Sech pulse creation"""
        duration = 100
        amp = 0.5
        sigma = 10
        # Zero ends = True
        Sech(duration=duration, amp=amp, sigma=sigma)

        # Zero ends = False
        Sech(duration=duration, amp=amp, sigma=sigma, zero_ends=False)

        with self.assertRaises(PulseError):
            Sech(duration=duration, amp=amp, sigma=-5)

    def test_sech_deriv_pulse(self):
        """Test that SechDeriv pulse creation"""
        duration = 100
        amp = 0.5
        sigma = 10
        SechDeriv(duration=duration, amp=amp, sigma=sigma)

        with self.assertRaises(PulseError):
            SechDeriv(duration=duration, amp=amp, sigma=-5)

    def test_constant_samples(self):
        """Test the constant pulse and its sampled construction."""
        amp = 0.6
        angle = np.pi * 0.7
        const = Constant(duration=150, amp=amp, angle=angle)
        self.assertEqual(const.get_waveform().samples[0], amp * np.exp(1j * angle))
        self.assertEqual(len(const.get_waveform().samples), 150)

    def test_parameters(self):
        """Test that the parameters can be extracted as a dict through the `parameters`
        attribute."""
        drag = Drag(duration=25, amp=0.2, sigma=7.8, beta=4, angle=0.2)
        self.assertEqual(set(drag.parameters.keys()), {"duration", "amp", "sigma", "beta", "angle"})
        const = Constant(duration=150, amp=1)
        self.assertEqual(set(const.parameters.keys()), {"duration", "amp", "angle"})

    def test_repr(self):
        """Test the repr methods for symbolic pulses."""
        gaus = Gaussian(duration=25, amp=0.7, sigma=4, angle=0.3)
        self.assertEqual(repr(gaus), "Gaussian(duration=25, sigma=4, amp=0.7, angle=0.3)")
        gaus_square = GaussianSquare(duration=20, sigma=30, amp=1.0, width=3)
        self.assertEqual(
            repr(gaus_square), "GaussianSquare(duration=20, sigma=30, width=3, amp=1.0, angle=0.0)"
        )
        gaus_square = GaussianSquare(
            duration=20, sigma=30, amp=1.0, angle=0.2, risefall_sigma_ratio=0.1
        )
        self.assertEqual(
            repr(gaus_square),
            "GaussianSquare(duration=20, sigma=30, width=14.0, amp=1.0, angle=0.2)",
        )
        gsd = GaussianSquareDrag(duration=20, sigma=30, amp=1.0, width=3, beta=1)
        self.assertEqual(
            repr(gsd),
            "GaussianSquareDrag(duration=20, sigma=30, width=3, beta=1, amp=1.0, angle=0.0)",
        )
        gsd = GaussianSquareDrag(duration=20, sigma=30, amp=1.0, risefall_sigma_ratio=0.1, beta=1)
        self.assertEqual(
            repr(gsd),
            "GaussianSquareDrag(duration=20, sigma=30, width=14.0, beta=1, amp=1.0, angle=0.0)",
        )
        gse = gaussian_square_echo(duration=20, sigma=30, amp=1.0, width=3)
        self.assertEqual(
            repr(gse),
            (
                "gaussian_square_echo(duration=20, amp=1.0, angle=0.0, sigma=30, width=3,"
                " active_amp=0.0, active_angle=0.0)"
            ),
        )
        gse = gaussian_square_echo(duration=20, sigma=30, amp=1.0, risefall_sigma_ratio=0.1)
        self.assertEqual(
            repr(gse),
            (
                "gaussian_square_echo(duration=20, amp=1.0, angle=0.0, sigma=30, width=14.0,"
                " active_amp=0.0, active_angle=0.0)"
            ),
        )
        drag = Drag(duration=5, amp=0.5, sigma=7, beta=1)
        self.assertEqual(repr(drag), "Drag(duration=5, sigma=7, beta=1, amp=0.5, angle=0.0)")
        const = Constant(duration=150, amp=0.1, angle=0.3)
        self.assertEqual(repr(const), "Constant(duration=150, amp=0.1, angle=0.3)")
        sin_pulse = Sin(duration=150, amp=0.1, angle=0.3, freq=0.2, phase=0)
        self.assertEqual(
            repr(sin_pulse), "Sin(duration=150, freq=0.2, phase=0, amp=0.1, angle=0.3)"
        )
        cos_pulse = Cos(duration=150, amp=0.1, angle=0.3, freq=0.2, phase=0)
        self.assertEqual(
            repr(cos_pulse), "Cos(duration=150, freq=0.2, phase=0, amp=0.1, angle=0.3)"
        )
        triangle_pulse = Triangle(duration=150, amp=0.1, angle=0.3, freq=0.2, phase=0)
        self.assertEqual(
            repr(triangle_pulse), "Triangle(duration=150, freq=0.2, phase=0, amp=0.1, angle=0.3)"
        )
        sawtooth_pulse = Sawtooth(duration=150, amp=0.1, angle=0.3, freq=0.2, phase=0)
        self.assertEqual(
            repr(sawtooth_pulse), "Sawtooth(duration=150, freq=0.2, phase=0, amp=0.1, angle=0.3)"
        )
        sech_pulse = Sech(duration=150, amp=0.1, angle=0.3, sigma=10)
        self.assertEqual(repr(sech_pulse), "Sech(duration=150, sigma=10, amp=0.1, angle=0.3)")
        sech_deriv_pulse = SechDeriv(duration=150, amp=0.1, angle=0.3, sigma=10)
        self.assertEqual(
            repr(sech_deriv_pulse), "SechDeriv(duration=150, sigma=10, amp=0.1, angle=0.3)"
        )
        gaussian_deriv_pulse = GaussianDeriv(duration=150, amp=0.1, angle=0.3, sigma=10)
        self.assertEqual(
            repr(gaussian_deriv_pulse), "GaussianDeriv(duration=150, sigma=10, amp=0.1, angle=0.3)"
        )

    def test_param_validation(self):
        """Test that symbolic pulse parameters are validated when initialized."""
        with self.assertRaises(PulseError):
            Gaussian(duration=25, sigma=0, amp=0.5, angle=np.pi / 2)
        with self.assertRaises(PulseError):
            GaussianSquare(duration=150, amp=0.2, sigma=8)
        with self.assertRaises(PulseError):
            GaussianSquare(duration=150, amp=0.2, sigma=8, width=100, risefall_sigma_ratio=5)
        with self.assertRaises(PulseError):
            GaussianSquare(duration=150, amp=0.2, sigma=8, width=160)
        with self.assertRaises(PulseError):
            GaussianSquare(duration=150, amp=0.2, sigma=8, risefall_sigma_ratio=10)

        with self.assertRaises(PulseError):
            GaussianSquareDrag(duration=150, amp=0.2, sigma=8, beta=1)
        with self.assertRaises(PulseError):
            GaussianSquareDrag(duration=150, amp=0.2, sigma=8, width=160, beta=1)
        with self.assertRaises(PulseError):
            GaussianSquareDrag(duration=150, amp=0.2, sigma=8, risefall_sigma_ratio=10, beta=1)

        with self.assertRaises(PulseError):
            gaussian_square_echo(
                duration=150,
                amp=0.2,
                sigma=8,
            )
        with self.assertRaises(PulseError):
            gaussian_square_echo(duration=150, amp=0.2, sigma=8, width=160)
        with self.assertRaises(PulseError):
            gaussian_square_echo(duration=150, amp=0.2, sigma=8, risefall_sigma_ratio=10)

        with self.assertRaises(PulseError):
            Constant(duration=150, amp=1.5, angle=np.pi * 0.8)
        with self.assertRaises(PulseError):
            Drag(duration=25, amp=0.5, sigma=-7.8, beta=4, angle=np.pi / 3)

    def test_class_level_limit_amplitude(self):
        """Test that the check for amplitude less than or equal to 1 can
        be disabled on the class level.

        Tests for representative examples.
        """
        with self.assertRaises(PulseError):
            Gaussian(duration=100, sigma=1.0, amp=1.7, angle=np.pi * 1.1)

        with patch("qiskit.pulse.library.pulse.Pulse.limit_amplitude", new=False):
            waveform = Gaussian(duration=100, sigma=1.0, amp=1.7, angle=np.pi * 1.1)
            self.assertGreater(np.abs(waveform.amp), 1.0)
            waveform = GaussianSquare(duration=100, sigma=1.0, amp=1.5, width=10, angle=np.pi / 5)
            self.assertGreater(np.abs(waveform.amp), 1.0)
            waveform = GaussianSquareDrag(duration=100, sigma=1.0, amp=1.1, beta=0.1, width=10)
            self.assertGreater(np.abs(waveform.amp), 1.0)

    def test_class_level_disable_validation(self):
        """Test that pulse validation can be disabled on the class level.

        Tests for representative examples.
        """
        with self.assertRaises(PulseError):
            Gaussian(duration=100, sigma=-1.0, amp=0.5, angle=np.pi * 1.1)

        with patch(
            "qiskit.pulse.library.symbolic_pulses.SymbolicPulse.disable_validation", new=True
        ):
            waveform = Gaussian(duration=100, sigma=-1.0, amp=0.5, angle=np.pi * 1.1)
            self.assertLess(waveform.sigma, 0)
            waveform = GaussianSquare(duration=100, sigma=1.0, amp=0.5, width=1000, angle=np.pi / 5)
            self.assertGreater(waveform.width, waveform.duration)
            waveform = GaussianSquareDrag(duration=100, sigma=1.0, amp=1.1, beta=0.1, width=-1)
            self.assertLess(waveform.width, 0)

    def test_gaussian_limit_amplitude_per_instance(self):
        """Test limit amplitude option per Gaussian instance."""
        with self.assertRaises(PulseError):
            Gaussian(duration=100, sigma=1.0, amp=1.6, angle=np.pi / 2.5)

        waveform = Gaussian(
            duration=100, sigma=1.0, amp=1.6, angle=np.pi / 2.5, limit_amplitude=False
        )
        self.assertGreater(np.abs(waveform.amp), 1.0)

    def test_gaussian_square_limit_amplitude_per_instance(self):
        """Test limit amplitude option per GaussianSquare instance."""
        with self.assertRaises(PulseError):
            GaussianSquare(duration=100, sigma=1.0, amp=1.5, width=10, angle=np.pi / 3)

        waveform = GaussianSquare(
            duration=100, sigma=1.0, amp=1.5, width=10, angle=np.pi / 3, limit_amplitude=False
        )
        self.assertGreater(np.abs(waveform.amp), 1.0)

    def test_gaussian_square_drag_limit_amplitude_per_instance(self):
        """Test limit amplitude option per GaussianSquareDrag instance."""
        with self.assertRaises(PulseError):
            GaussianSquareDrag(duration=100, sigma=1.0, amp=1.1, beta=0.1, width=10)

        waveform = GaussianSquareDrag(
            duration=100, sigma=1.0, amp=1.1, beta=0.1, width=10, limit_amplitude=False
        )
        self.assertGreater(np.abs(waveform.amp), 1.0)

    def test_gaussian_square_echo_limit_amplitude_per_instance(self):
        """Test limit amplitude option per GaussianSquareEcho instance."""
        with self.assertRaises(PulseError):
            gaussian_square_echo(duration=1000, sigma=4.0, amp=1.01, width=100)

        waveform = gaussian_square_echo(
            duration=1000, sigma=4.0, amp=1.01, width=100, limit_amplitude=False
        )
        self.assertGreater(np.abs(waveform.amp), 1.0)

    def test_drag_limit_amplitude_per_instance(self):
        """Test limit amplitude option per DRAG instance."""
        with self.assertRaises(PulseError):
            Drag(duration=100, sigma=1.0, beta=1.0, amp=1.8, angle=np.pi * 0.3)

        waveform = Drag(
            duration=100, sigma=1.0, beta=1.0, amp=1.8, angle=np.pi * 0.3, limit_amplitude=False
        )
        self.assertGreater(np.abs(waveform.amp), 1.0)

    def test_constant_limit_amplitude_per_instance(self):
        """Test limit amplitude option per Constant instance."""
        with self.assertRaises(PulseError):
            Constant(duration=100, amp=1.6, angle=0.5)

        waveform = Constant(duration=100, amp=1.6, angle=0.5, limit_amplitude=False)
        self.assertGreater(np.abs(waveform.amp), 1.0)

    def test_sin_limit_amplitude_per_instance(self):
        """Test limit amplitude option per Sin instance."""
        with self.assertRaises(PulseError):
            Sin(duration=100, amp=1.1, phase=0)

        waveform = Sin(duration=100, amp=1.1, phase=0, limit_amplitude=False)
        self.assertGreater(np.abs(waveform.amp), 1.0)

    def test_sawtooth_limit_amplitude_per_instance(self):
        """Test limit amplitude option per Sawtooth instance."""
        with self.assertRaises(PulseError):
            Sawtooth(duration=100, amp=1.1, phase=0)

        waveform = Sawtooth(duration=100, amp=1.1, phase=0, limit_amplitude=False)
        self.assertGreater(np.abs(waveform.amp), 1.0)

    def test_triangle_limit_amplitude_per_instance(self):
        """Test limit amplitude option per Triangle instance."""
        with self.assertRaises(PulseError):
            Triangle(duration=100, amp=1.1, phase=0)

        waveform = Triangle(duration=100, amp=1.1, phase=0, limit_amplitude=False)
        self.assertGreater(np.abs(waveform.amp), 1.0)

    def test_square_limit_amplitude_per_instance(self):
        """Test limit amplitude option per Square instance."""
        with self.assertRaises(PulseError):
            Square(duration=100, amp=1.1, phase=0)

        waveform = Square(duration=100, amp=1.1, phase=0, limit_amplitude=False)
        self.assertGreater(np.abs(waveform.amp), 1.0)

    def test_gaussian_deriv_limit_amplitude_per_instance(self):
        """Test limit amplitude option per GaussianDeriv instance."""
        with self.assertRaises(PulseError):
            GaussianDeriv(duration=100, amp=5, sigma=1)

        waveform = GaussianDeriv(duration=100, amp=5, sigma=1, limit_amplitude=False)
        self.assertGreater(np.abs(waveform.amp / waveform.sigma), np.exp(0.5))

    def test_sech_limit_amplitude_per_instance(self):
        """Test limit amplitude option per Sech instance."""
        with self.assertRaises(PulseError):
            Sech(duration=100, amp=5, sigma=1)

        waveform = Sech(duration=100, amp=5, sigma=1, limit_amplitude=False)
        self.assertGreater(np.abs(waveform.amp), 1.0)

    def test_sech_deriv_limit_amplitude_per_instance(self):
        """Test limit amplitude option per SechDeriv instance."""
        with self.assertRaises(PulseError):
            SechDeriv(duration=100, amp=5, sigma=1)

        waveform = SechDeriv(duration=100, amp=5, sigma=1, limit_amplitude=False)
        self.assertGreater(np.abs(waveform.amp) / waveform.sigma, 2.0)

    def test_get_parameters(self):
        """Test getting pulse parameters as attribute."""
        drag_pulse = Drag(duration=100, amp=0.1, sigma=40, beta=3)
        self.assertEqual(drag_pulse.duration, 100)
        self.assertEqual(drag_pulse.amp, 0.1)
        self.assertEqual(drag_pulse.sigma, 40)
        self.assertEqual(drag_pulse.beta, 3)

        with self.assertRaises(AttributeError):
            _ = drag_pulse.non_existing_parameter

    def test_envelope_cache(self):
        """Test speed up of instantiation with lambdify envelope cache."""
        drag_instance1 = Drag(duration=100, amp=0.1, sigma=40, beta=3)
        drag_instance2 = Drag(duration=100, amp=0.1, sigma=40, beta=3)
        self.assertTrue(drag_instance1._envelope_lam is drag_instance2._envelope_lam)

    def test_constraints_cache(self):
        """Test speed up of instantiation with lambdify constraints cache."""
        drag_instance1 = Drag(duration=100, amp=0.1, sigma=40, beta=3)
        drag_instance2 = Drag(duration=100, amp=0.1, sigma=40, beta=3)
        self.assertTrue(drag_instance1._constraints_lam is drag_instance2._constraints_lam)

    def test_deepcopy(self):
        """Test deep copying instance."""
        import copy

        drag = Drag(duration=100, amp=0.1, sigma=40, beta=3)
        drag_copied = copy.deepcopy(drag)

        self.assertNotEqual(id(drag), id(drag_copied))

        orig_wf = drag.get_waveform()
        copied_wf = drag_copied.get_waveform()

        np.testing.assert_almost_equal(orig_wf.samples, copied_wf.samples)

    def test_fully_parametrized_pulse(self):
        """Test instantiating a pulse with parameters."""
        amp = Parameter("amp")
        duration = Parameter("duration")
        sigma = Parameter("sigma")
        beta = Parameter("beta")

        # doesn't raise an error
        drag = Drag(duration=duration, amp=amp, sigma=sigma, beta=beta)

        with self.assertRaises(PulseError):
            drag.get_waveform()

    # pylint: disable=invalid-name
    def test_custom_pulse(self):
        """Test defining a custom pulse which is not in the form of amp * F(t)."""
        t, t1, t2, amp1, amp2 = sym.symbols("t, t1, t2, amp1, amp2")
        envelope = sym.Piecewise((amp1, sym.And(t > t1, t < t2)), (amp2, sym.true))

        custom_pulse = SymbolicPulse(
            pulse_type="Custom",
            duration=100,
            parameters={"t1": 30, "t2": 80, "amp1": 0.1j, "amp2": -0.1},
            envelope=envelope,
        )
        waveform = custom_pulse.get_waveform()
        reference = np.concatenate([-0.1 * np.ones(30), 0.1j * np.ones(50), -0.1 * np.ones(20)])
        np.testing.assert_array_almost_equal(waveform.samples, reference)

    def test_gaussian_deprecated_type_check(self):
        """Test isinstance check works with deprecation."""
        gaussian_pulse = Gaussian(160, 0.1, 40)

        self.assertTrue(isinstance(gaussian_pulse, SymbolicPulse))
        self.assertTrue(isinstance(gaussian_pulse, Gaussian))
        self.assertFalse(isinstance(gaussian_pulse, GaussianSquare))
        self.assertFalse(isinstance(gaussian_pulse, Drag))
        self.assertFalse(isinstance(gaussian_pulse, Constant))

    def test_gaussian_square_deprecated_type_check(self):
        """Test isinstance check works with deprecation."""
        gaussian_square_pulse = GaussianSquare(800, 0.1, 64, 544)

        self.assertTrue(isinstance(gaussian_square_pulse, SymbolicPulse))
        self.assertFalse(isinstance(gaussian_square_pulse, Gaussian))
        self.assertTrue(isinstance(gaussian_square_pulse, GaussianSquare))
        self.assertFalse(isinstance(gaussian_square_pulse, Drag))
        self.assertFalse(isinstance(gaussian_square_pulse, Constant))

    def test_drag_deprecated_type_check(self):
        """Test isinstance check works with deprecation."""
        drag_pulse = Drag(160, 0.1, 40, 1.5)

        self.assertTrue(isinstance(drag_pulse, SymbolicPulse))
        self.assertFalse(isinstance(drag_pulse, Gaussian))
        self.assertFalse(isinstance(drag_pulse, GaussianSquare))
        self.assertTrue(isinstance(drag_pulse, Drag))
        self.assertFalse(isinstance(drag_pulse, Constant))

    def test_constant_deprecated_type_check(self):
        """Test isinstance check works with deprecation."""
        constant_pulse = Constant(160, 0.1, 40, 1.5)

        self.assertTrue(isinstance(constant_pulse, SymbolicPulse))
        self.assertFalse(isinstance(constant_pulse, Gaussian))
        self.assertFalse(isinstance(constant_pulse, GaussianSquare))
        self.assertFalse(isinstance(constant_pulse, Drag))
        self.assertTrue(isinstance(constant_pulse, Constant))


@decorate_test_methods(ignore_pulse_deprecation_warnings)
class TestFunctionalPulse(QiskitTestCase):
    """Waveform tests."""

    # pylint: disable=invalid-name
    def test_gaussian(self):
        """Test gaussian pulse."""

        @functional_pulse
        def local_gaussian(duration, amp, t0, sig):
            x = np.linspace(0, duration - 1, duration)
            return amp * np.exp(-((x - t0) ** 2) / sig**2)

        pulse_wf_inst = local_gaussian(duration=10, amp=1, t0=5, sig=1, name="test_pulse")
        _y = 1 * np.exp(-((np.linspace(0, 9, 10) - 5) ** 2) / 1**2)

        self.assertListEqual(list(pulse_wf_inst.samples), list(_y))

        # check name
        self.assertEqual(pulse_wf_inst.name, "test_pulse")

        # check duration
        self.assertEqual(pulse_wf_inst.duration, 10)

    # pylint: disable=invalid-name
    def test_variable_duration(self):
        """Test generation of sample pulse with variable duration."""

        @functional_pulse
        def local_gaussian(duration, amp, t0, sig):
            x = np.linspace(0, duration - 1, duration)
            return amp * np.exp(-((x - t0) ** 2) / sig**2)

        _durations = np.arange(10, 15, 1)

        for _duration in _durations:
            pulse_wf_inst = local_gaussian(duration=_duration, amp=1, t0=5, sig=1)
            self.assertEqual(len(pulse_wf_inst.samples), _duration)


@decorate_test_methods(ignore_pulse_deprecation_warnings)
class TestScalableSymbolicPulse(QiskitTestCase):
    """ScalableSymbolicPulse tests"""

    def test_scalable_comparison(self):
        """Test equating of pulses"""
        # amp,angle comparison
        gaussian_negamp = Gaussian(duration=25, sigma=4, amp=-0.5, angle=0)
        gaussian_piphase = Gaussian(duration=25, sigma=4, amp=0.5, angle=np.pi)
        self.assertEqual(gaussian_negamp, gaussian_piphase)

        # Parameterized library pulses
        amp = Parameter("amp")
        gaussian1 = Gaussian(duration=25, sigma=4, amp=amp, angle=0)
        gaussian2 = Gaussian(duration=25, sigma=4, amp=amp, angle=0)
        self.assertEqual(gaussian1, gaussian2)

        # pulses with different parameters
        gaussian1._params["sigma"] = 10
        self.assertNotEqual(gaussian1, gaussian2)

    def test_complex_amp_error(self):
        """Test that initializing a pulse with complex amp raises an error"""
        with self.assertRaises(PulseError):
            ScalableSymbolicPulse("test", duration=100, amp=0.1j, angle=0.0)


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


"""Tests for Stabilizerstate quantum state class."""

from itertools import product
import unittest
import logging
from ddt import ddt, data, unpack

import numpy as np

from qiskit import QuantumCircuit

from qiskit.quantum_info.random import random_clifford, random_pauli
from qiskit.quantum_info.states import StabilizerState, Statevector
from qiskit.circuit.library import IGate, XGate, HGate
from qiskit.quantum_info.operators import Clifford, Pauli, Operator
from test import combine  # pylint: disable=wrong-import-order
from test import QiskitTestCase  # pylint: disable=wrong-import-order


logger = logging.getLogger(__name__)


class StabilizerStateTestingTools:
    """Test tools for verifying test cases in StabilizerState"""

    @staticmethod
    def _bitstring_product_dict(bitstring_length: int, skip_entries: dict = None) -> dict:
        """Retrieves a dict of every possible product of '0', '1' for length bitstring_length
        pass in a dict to use the keys as entries to skip adding to the dict

        Args:
            bitstring_length (int): length of the bitstring product
            skip_entries (dict[str, float], optional): dict entries to skip adding to the dict based
                on existing keys in the dict passed in. Defaults to {}.

        Returns:
            dict[str, float]: dict with entries, all set to 0
        """
        if skip_entries is None:
            skip_entries = {}
        return {
            result: 0
            for result in ["".join(x) for x in product(["0", "1"], repeat=bitstring_length)]
            if result not in skip_entries
        }

    @staticmethod
    def _verify_individual_bitstrings(
        testcase: QiskitTestCase,
        target_dict: dict,
        stab: StabilizerState,
        qargs: list = None,
        decimals: int = None,
        dict_almost_equal: bool = False,
    ) -> None:
        """Helper that iterates through the target_dict and checks all probabilities by
        running the value through the probabilities_dict_from_bitstring method for
        retrieving a single measurement

        Args:
            target_dict (dict[str, float]): dict to check probabilities for
            stab (StabilizerState): stabilizerstate object to run probabilities_dict_from_bitstring on
            qargs (None or list): subsystems to return probabilities for,
                    if None return for all subsystems (Default: None).
            decimals (None or int): the number of decimal places to round
                    values. If None no rounding is done (Default: None)
            dict_almost_equal (bool): utilize assertDictAlmostEqual when true, assertDictEqual when false
        """
        for outcome_bitstring in target_dict:
            (testcase.assertDictAlmostEqual if (dict_almost_equal) else testcase.assertDictEqual)(
                stab.probabilities_dict_from_bitstring(
                    outcome_bitstring=outcome_bitstring, qargs=qargs, decimals=decimals
                ),
                {outcome_bitstring: target_dict[outcome_bitstring]},
            )


@ddt
class TestStabilizerState(QiskitTestCase):
    """Tests for StabilizerState class."""

    rng = np.random.default_rng(12345)
    samples = 10
    shots = 1000
    threshold = 0.1 * shots

    @combine(num_qubits=[2, 3, 4, 5])
    def test_init_clifford(self, num_qubits):
        """Test initialization from Clifford."""
        stab1 = StabilizerState(random_clifford(num_qubits, seed=self.rng))
        stab2 = StabilizerState(stab1)
        self.assertEqual(stab1, stab2)

    @combine(num_qubits=[2, 3, 4, 5])
    def test_init_circuit(self, num_qubits):
        """Test initialization from a Clifford circuit."""
        cliff = random_clifford(num_qubits, seed=self.rng)
        stab1 = StabilizerState(cliff.to_circuit())
        stab2 = StabilizerState(cliff)
        self.assertEqual(stab1, stab2)

    @combine(num_qubits=[2, 3, 4, 5])
    def test_init_instruction(self, num_qubits):
        """Test initialization from a Clifford instruction."""
        cliff = random_clifford(num_qubits, seed=self.rng)
        stab1 = StabilizerState(cliff.to_instruction())
        stab2 = StabilizerState(cliff)
        self.assertEqual(stab1, stab2)

    @combine(num_qubits=[2, 3, 4, 5])
    def test_init_pauli(self, num_qubits):
        """Test initialization from pauli."""
        pauli = random_pauli(num_qubits, seed=self.rng)
        stab1 = StabilizerState(pauli)
        stab2 = StabilizerState(stab1)
        self.assertEqual(stab1, stab2)

    @combine(num_qubits=[2, 3, 4, 5])
    def test_to_operator(self, num_qubits):
        """Test to_operator method for returning projector."""
        for _ in range(self.samples):
            stab = StabilizerState(random_clifford(num_qubits, seed=self.rng))
            target = Operator(stab)
            op = StabilizerState(stab).to_operator()
            self.assertEqual(op, target)

    @combine(num_qubits=[2, 3, 4])
    def test_trace(self, num_qubits):
        """Test trace methods"""
        stab = StabilizerState(random_clifford(num_qubits, seed=self.rng))
        trace = stab.trace()
        self.assertEqual(trace, 1.0)

    @combine(num_qubits=[2, 3, 4])
    def test_purity(self, num_qubits):
        """Test purity methods"""
        stab = StabilizerState(random_clifford(num_qubits, seed=self.rng))
        purity = stab.purity()
        self.assertEqual(purity, 1.0)

    @combine(num_qubits=[2, 3])
    def test_conjugate(self, num_qubits):
        """Test conjugate method."""
        for _ in range(self.samples):
            cliff = random_clifford(num_qubits, seed=self.rng)
            target = StabilizerState(cliff.conjugate())
            state = StabilizerState(cliff).conjugate()
            self.assertEqual(state, target)

    def test_tensor(self):
        """Test tensor method."""
        for _ in range(self.samples):
            cliff1 = random_clifford(2, seed=self.rng)
            cliff2 = random_clifford(3, seed=self.rng)
            stab1 = StabilizerState(cliff1)
            stab2 = StabilizerState(cliff2)
            target = StabilizerState(cliff1.tensor(cliff2))
            state = stab1.tensor(stab2)
            self.assertEqual(state, target)

    def test_expand(self):
        """Test expand method."""
        for _ in range(self.samples):
            cliff1 = random_clifford(2, seed=self.rng)
            cliff2 = random_clifford(3, seed=self.rng)
            stab1 = StabilizerState(cliff1)
            stab2 = StabilizerState(cliff2)
            target = StabilizerState(cliff1.expand(cliff2))
            state = stab1.expand(stab2)
            self.assertEqual(state, target)

    @combine(num_qubits=[2, 3, 4])
    def test_evolve(self, num_qubits):
        """Test evolve method."""
        for _ in range(self.samples):
            cliff1 = random_clifford(num_qubits, seed=self.rng)
            cliff2 = random_clifford(num_qubits, seed=self.rng)
            stab1 = StabilizerState(cliff1)
            stab2 = StabilizerState(cliff2)
            target = StabilizerState(cliff1.compose(cliff2))
            state = stab1.evolve(stab2)
            self.assertEqual(state, target)

    @combine(num_qubits_1=[4, 5, 6], num_qubits_2=[1, 2, 3])
    def test_evolve_subsystem(self, num_qubits_1, num_qubits_2):
        """Test subsystem evolve method."""
        for _ in range(self.samples):
            cliff1 = random_clifford(num_qubits_1, seed=self.rng)
            cliff2 = random_clifford(num_qubits_2, seed=self.rng)
            stab1 = StabilizerState(cliff1)
            stab2 = StabilizerState(cliff2)
            qargs = sorted(np.random.choice(range(num_qubits_1), num_qubits_2, replace=False))
            target = StabilizerState(cliff1.compose(cliff2, qargs))
            state = stab1.evolve(stab2, qargs)
            self.assertEqual(state, target)

    def test_measure_single_qubit(self):
        """Test a measurement of a single qubit"""
        for _ in range(self.samples):
            cliff = Clifford(XGate())
            stab = StabilizerState(cliff)
            value = stab.measure()[0]
            self.assertEqual(value, "1")

            cliff = Clifford(IGate())
            stab = StabilizerState(cliff)
            value = stab.measure()[0]
            self.assertEqual(value, "0")

            cliff = Clifford(HGate())
            stab = StabilizerState(cliff)
            value = stab.measure()[0]
            self.assertIn(value, ["0", "1"])

    def test_measure_qubits(self):
        """Test a measurement of a subsystem of qubits"""

        for _ in range(self.samples):
            num_qubits = 4
            qc = QuantumCircuit(num_qubits)
            stab = StabilizerState(qc)
            value = stab.measure()[0]
            self.assertEqual(value, "0000")
            value = stab.measure([0, 2])[0]
            self.assertEqual(value, "00")
            value = stab.measure([1])[0]
            self.assertEqual(value, "0")

            for i in range(num_qubits):
                qc.x(i)
            stab = StabilizerState(qc)
            value = stab.measure()[0]
            self.assertEqual(value, "1111")
            value = stab.measure([2, 0])[0]
            self.assertEqual(value, "11")
            value = stab.measure([1])[0]
            self.assertEqual(value, "1")

            qc = QuantumCircuit(num_qubits)
            qc.h(0)
            stab = StabilizerState(qc)
            value = stab.measure()[0]
            self.assertIn(value, ["0000", "0001"])
            value = stab.measure([0, 1])[0]
            self.assertIn(value, ["00", "01"])
            value = stab.measure([2])[0]
            self.assertEqual(value, "0")

            qc = QuantumCircuit(num_qubits)
            qc.h(0)
            qc.cx(0, 1)
            qc.cx(0, 2)
            qc.cx(0, 3)
            stab = StabilizerState(qc)
            value = stab.measure()[0]
            self.assertIn(value, ["0000", "1111"])
            value = stab.measure([3, 1])[0]
            self.assertIn(value, ["00", "11"])
            value = stab.measure([2])[0]
            self.assertIn(value, ["0", "1"])

    def test_reset_single_qubit(self):
        """Test reset method of a single qubit"""

        empty_qc = QuantumCircuit(1)

        for _ in range(self.samples):
            cliff = Clifford(XGate())
            stab = StabilizerState(cliff)
            value = stab.reset([0])
            target = StabilizerState(empty_qc)
            self.assertEqual(value, target)

            cliff = Clifford(HGate())
            stab = StabilizerState(cliff)
            value = stab.reset([0])
            target = StabilizerState(empty_qc)
            self.assertEqual(value, target)

    def test_reset_qubits(self):
        """Test reset method of a subsystem of qubits"""

        num_qubits = 3
        qc = QuantumCircuit(num_qubits)
        qc.h(0)
        qc.cx(0, 1)
        qc.cx(0, 2)

        for _ in range(self.samples):
            with self.subTest(msg="reset (None)"):
                stab = StabilizerState(qc)
                res = stab.reset()
                value = res.measure()[0]
                self.assertEqual(value, "000")

        for _ in range(self.samples):
            for qargs in [[0, 1, 2], [2, 1, 0], [1, 2, 0], [1, 0, 2]]:
                with self.subTest(msg=f"reset (qargs={qargs})"):
                    stab = StabilizerState(qc)
                    res = stab.reset(qargs)
                    value = res.measure()[0]
                    self.assertEqual(value, "000")

        for _ in range(self.samples):
            with self.subTest(msg="reset ([0])"):
                stab = StabilizerState(qc)
                res = stab.reset([0])
                value = res.measure()[0]
                self.assertIn(value, ["000", "110"])

        for _ in range(self.samples):
            with self.subTest(msg="reset ([1])"):
                stab = StabilizerState(qc)
                res = stab.reset([1])
                value = res.measure()[0]
                self.assertIn(value, ["000", "101"])

        for _ in range(self.samples):
            with self.subTest(msg="reset ([2])"):
                stab = StabilizerState(qc)
                res = stab.reset([2])
                value = res.measure()[0]
                self.assertIn(value, ["000", "011"])

        for _ in range(self.samples):
            for qargs in [[0, 1], [1, 0]]:
                with self.subTest(msg=f"reset (qargs={qargs})"):
                    stab = StabilizerState(qc)
                    res = stab.reset(qargs)
                    value = res.measure()[0]
                    self.assertIn(value, ["000", "100"])

        for _ in range(self.samples):
            for qargs in [[0, 2], [2, 0]]:
                with self.subTest(msg=f"reset (qargs={qargs})"):
                    stab = StabilizerState(qc)
                    res = stab.reset(qargs)
                    value = res.measure()[0]
                    self.assertIn(value, ["000", "010"])

        for _ in range(self.samples):
            for qargs in [[1, 2], [2, 1]]:
                with self.subTest(msg=f"reset (qargs={qargs})"):
                    stab = StabilizerState(qc)
                    res = stab.reset(qargs)
                    value = res.measure()[0]
                    self.assertIn(value, ["000", "001"])

    def test_probabilities_dict_single_qubit(self):
        """Test probabilities and probabilities_dict methods of a single qubit"""

        num_qubits = 1
        qc = QuantumCircuit(num_qubits)

        for _ in range(self.samples):
            with self.subTest(msg="P(id(0))"):
                stab = StabilizerState(qc)
                value = stab.probabilities_dict()
                target = {"0": 1}
                self.assertEqual(value, target)
                target.update({"1": 0.0})
                StabilizerStateTestingTools._verify_individual_bitstrings(self, target, stab)
                probs = stab.probabilities()
                target = np.array([1, 0])
                self.assertTrue(np.allclose(probs, target))

        qc.x(0)
        for _ in range(self.samples):
            with self.subTest(msg="P(x(0))"):
                stab = StabilizerState(qc)
                value = stab.probabilities_dict()
                target = {"1": 1}
                self.assertEqual(value, target)
                target.update({"0": 0.0})
                StabilizerStateTestingTools._verify_individual_bitstrings(self, target, stab)
                probs = stab.probabilities()
                target = np.array([0, 1])
                self.assertTrue(np.allclose(probs, target))

        qc = QuantumCircuit(num_qubits)
        qc.h(0)
        for _ in range(self.samples):
            with self.subTest(msg="P(h(0))"):
                stab = StabilizerState(qc)
                value = stab.probabilities_dict()
                target = {"0": 0.5, "1": 0.5}
                self.assertEqual(value, target)
                StabilizerStateTestingTools._verify_individual_bitstrings(self, target, stab)
                probs = stab.probabilities()
                target = np.array([0.5, 0.5])
                self.assertTrue(np.allclose(probs, target))

    def test_probabilities_dict_two_qubits(self):
        """Test probabilities and probabilities_dict methods of two qubits"""

        num_qubits = 2
        qc = QuantumCircuit(num_qubits)
        qc.h(0)
        stab = StabilizerState(qc)

        for _ in range(self.samples):
            with self.subTest(msg="P(None)"):
                value = stab.probabilities_dict()
                target = {"00": 0.5, "01": 0.5}
                self.assertEqual(value, target)
                target.update({"10": 0.0, "11": 0.0})
                StabilizerStateTestingTools._verify_individual_bitstrings(self, target, stab)
                probs = stab.probabilities()
                target = np.array([0.5, 0.5, 0, 0])
                self.assertTrue(np.allclose(probs, target))

        qargs: list = [0, 1]
        for _ in range(self.samples):
            with self.subTest(msg="P([0, 1])"):
                value = stab.probabilities_dict(qargs)
                target = {"00": 0.5, "01": 0.5}
                self.assertEqual(value, target)
                target.update({"10": 0.0, "11": 0.0})
                StabilizerStateTestingTools._verify_individual_bitstrings(self, target, stab, qargs)
                probs = stab.probabilities(qargs)
                target = np.array([0.5, 0.5, 0, 0])
                self.assertTrue(np.allclose(probs, target))

        qargs: list = [1, 0]
        for _ in range(self.samples):
            with self.subTest(msg="P([1, 0])"):
                value = stab.probabilities_dict(qargs)
                target = {"00": 0.5, "10": 0.5}
                self.assertEqual(value, target)
                target.update({"01": 0.0, "11": 0.0})
                StabilizerStateTestingTools._verify_individual_bitstrings(self, target, stab, qargs)
                probs = stab.probabilities(qargs)
                target = np.array([0.5, 0, 0.5, 0])
                self.assertTrue(np.allclose(probs, target))

        qargs: list = [0]
        for _ in range(self.samples):
            with self.subTest(msg="P[0]"):
                value = stab.probabilities_dict(qargs)
                target = {"0": 0.5, "1": 0.5}
                self.assertEqual(value, target)
                StabilizerStateTestingTools._verify_individual_bitstrings(self, target, stab, qargs)
                probs = stab.probabilities(qargs)
                target = np.array([0.5, 0.5])
                self.assertTrue(np.allclose(probs, target))

        qargs: list = [1]
        for _ in range(self.samples):
            with self.subTest(msg="P([1])"):
                value = stab.probabilities_dict(qargs)
                target = {"0": 1.0}
                self.assertEqual(value, target)
                target.update({"1": 0.0})
                StabilizerStateTestingTools._verify_individual_bitstrings(self, target, stab, qargs)
                probs = stab.probabilities(qargs)
                target = np.array([1, 0])
                self.assertTrue(np.allclose(probs, target))

    def test_probabilities_dict_qubits(self):
        """Test probabilities and probabilities_dict methods of a subsystem of qubits"""

        num_qubits = 3
        qc = QuantumCircuit(num_qubits)
        qc.h(0)
        qc.h(1)
        qc.h(2)
        stab = StabilizerState(qc)

        decimals: int = 1
        for _ in range(self.samples):
            with self.subTest(msg="P(None), decimals=1"):
                value = stab.probabilities_dict(decimals=decimals)
                target = {
                    "000": 0.1,
                    "001": 0.1,
                    "010": 0.1,
                    "011": 0.1,
                    "100": 0.1,
                    "101": 0.1,
                    "110": 0.1,
                    "111": 0.1,
                }
                self.assertEqual(value, target)
                StabilizerStateTestingTools._verify_individual_bitstrings(
                    self, target, stab, decimals=decimals
                )
                probs = stab.probabilities(decimals=decimals)
                target = np.array([0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1])
                self.assertTrue(np.allclose(probs, target))

        decimals: int = 2
        for _ in range(self.samples):
            with self.subTest(msg="P(None), decimals=2"):
                value = stab.probabilities_dict(decimals=decimals)
                target = {
                    "000": 0.12,
                    "001": 0.12,
                    "010": 0.12,
                    "011": 0.12,
                    "100": 0.12,
                    "101": 0.12,
                    "110": 0.12,
                    "111": 0.12,
                }
                self.assertEqual(value, target)
                StabilizerStateTestingTools._verify_individual_bitstrings(
                    self, target, stab, decimals=decimals
                )
                probs = stab.probabilities(decimals=decimals)
                target = np.array([0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12])
                self.assertTrue(np.allclose(probs, target))

        decimals: int = 3
        for _ in range(self.samples):
            with self.subTest(msg="P(None), decimals=3"):
                value = stab.probabilities_dict(decimals=decimals)
                target = {
                    "000": 0.125,
                    "001": 0.125,
                    "010": 0.125,
                    "011": 0.125,
                    "100": 0.125,
                    "101": 0.125,
                    "110": 0.125,
                    "111": 0.125,
                }
                self.assertEqual(value, target)
                StabilizerStateTestingTools._verify_individual_bitstrings(
                    self, target, stab, decimals=decimals
                )
                probs = stab.probabilities(decimals=3)
                target = np.array([0.125, 0.125, 0.125, 0.125, 0.125, 0.125, 0.125, 0.125])
                self.assertTrue(np.allclose(probs, target))

    @combine(num_qubits=[5, 6, 7, 8, 9])
    def test_probabilities_dict_from_bitstring(self, num_qubits):
        """Test probabilities_dict_from_bitstring methods with medium number of qubits that are still
        reasonable to calculate the full dict with probabilities_dict of all possible outcomes"""

        qc: QuantumCircuit = QuantumCircuit(num_qubits)
        for qubit_num in range(0, num_qubits):
            qc.h(qubit_num)
        stab = StabilizerState(qc)

        expected_result: float = float(1 / (2**num_qubits))
        target_dict: dict = StabilizerStateTestingTools._bitstring_product_dict(num_qubits)
        target_dict.update((k, expected_result) for k in target_dict)

        for _ in range(self.samples):
            with self.subTest(msg="P(None)"):
                value = stab.probabilities_dict()
                self.assertDictEqual(value, target_dict)
                StabilizerStateTestingTools._verify_individual_bitstrings(self, target_dict, stab)
                probs = stab.probabilities()
                target = np.array(([expected_result] * (2**num_qubits)))
                self.assertTrue(np.allclose(probs, target))

        # H gate at qubit 0, Every gate after is an X gate
        # will result in 2 outcomes with 0.5
        qc = QuantumCircuit(num_qubits)
        qc.h(0)
        for qubit_num in range(1, num_qubits):
            qc.x(qubit_num)
        stab = StabilizerState(qc)

        # Build the 2 expected outcome bitstrings for
        # 0.5 probability based on h and x gates
        target_1: str = "".join(["1" * (num_qubits - 1)] + ["0"])
        target_2: str = "".join(["1" * num_qubits])
        target: dict = {target_1: 0.5, target_2: 0.5}
        target_all_bitstrings: dict = StabilizerStateTestingTools._bitstring_product_dict(
            num_qubits, target
        )
        target_all_bitstrings.update(target_all_bitstrings)

        # Numpy Array to verify stab.probabilities()
        target_np_dict: dict = StabilizerStateTestingTools._bitstring_product_dict(
            num_qubits, [target_1, target_2]
        )
        target_np_dict.update(target)
        target_np_array: np.ndarray = np.array(list(target_np_dict.values()))

        for _ in range(self.samples):
            with self.subTest(msg="P(None)"):
                stab = StabilizerState(qc)
                value = stab.probabilities_dict()
                self.assertEqual(value, target)
                StabilizerStateTestingTools._verify_individual_bitstrings(
                    self, target_all_bitstrings, stab
                )
                probs = stab.probabilities()
                self.assertTrue(np.allclose(probs, target_np_array))

    def test_probabilities_dict_ghz(self):
        """Test probabilities and probabilities_dict method of a subsystem of qubits"""

        num_qubits = 3
        qc = QuantumCircuit(num_qubits)
        qc.h(0)
        qc.cx(0, 1)
        qc.cx(0, 2)

        with self.subTest(msg="P(None)"):
            stab = StabilizerState(qc)
            value = stab.probabilities_dict()
            target = {"000": 0.5, "111": 0.5}
            self.assertEqual(value, target)
            target.update(StabilizerStateTestingTools._bitstring_product_dict(num_qubits, target))
            StabilizerStateTestingTools._verify_individual_bitstrings(self, target, stab)
            probs = stab.probabilities()
            target = np.array([0.5, 0, 0, 0, 0, 0, 0, 0.5])
            self.assertTrue(np.allclose(probs, target))

        # 3-qubit qargs
        for qargs in [[0, 1, 2], [2, 1, 0], [1, 2, 0], [1, 0, 2]]:
            with self.subTest(msg=f"P({qargs})"):
                probs = stab.probabilities_dict(qargs)
                target = {"000": 0.5, "111": 0.5}
                self.assertDictAlmostEqual(probs, target)
                target.update(
                    StabilizerStateTestingTools._bitstring_product_dict(num_qubits, target)
                )
                StabilizerStateTestingTools._verify_individual_bitstrings(self, target, stab, qargs)
                probs = stab.probabilities(qargs)
                target = np.array([0.5, 0, 0, 0, 0, 0, 0, 0.5])
                self.assertTrue(np.allclose(probs, target))

        # 2-qubit qargs
        for qargs in [[0, 1], [2, 1], [1, 0], [1, 2]]:
            with self.subTest(msg=f"P({qargs})"):
                probs = stab.probabilities_dict(qargs)
                target = {"00": 0.5, "11": 0.5}
                self.assertDictAlmostEqual(probs, target)
                target.update(StabilizerStateTestingTools._bitstring_product_dict(2, target))
                StabilizerStateTestingTools._verify_individual_bitstrings(
                    self, target, stab, qargs, dict_almost_equal=True
                )
                probs = stab.probabilities(qargs)
                target = np.array([0.5, 0, 0, 0.5])
                self.assertTrue(np.allclose(probs, target))

        # 1-qubit qargs
        for qargs in [[0], [1], [2]]:
            with self.subTest(msg=f"P({qargs})"):
                probs = stab.probabilities_dict(qargs)
                target = {"0": 0.5, "1": 0.5}
                self.assertDictAlmostEqual(probs, target)
                StabilizerStateTestingTools._verify_individual_bitstrings(
                    self, target, stab, qargs, dict_almost_equal=True
                )
                probs = stab.probabilities(qargs)
                target = np.array([0.5, 0.5])
                self.assertTrue(np.allclose(probs, target))

    @combine(num_qubits=[2, 3, 4])
    def test_probs_random_subsystem(self, num_qubits):
        """Test probabilities and probabilities_dict methods
        of random cliffords for a subsystem of qubits"""

        for _ in range(self.samples):
            for subsystem_size in range(1, num_qubits):
                cliff = random_clifford(num_qubits, seed=self.rng)
                qargs = np.random.choice(num_qubits, size=subsystem_size, replace=False)
                qc = cliff.to_circuit()
                stab = StabilizerState(cliff)
                probs = stab.probabilities(qargs)
                probs_dict = stab.probabilities_dict(qargs)
                StabilizerStateTestingTools._verify_individual_bitstrings(
                    self, probs_dict, stab, qargs
                )
                target = Statevector(qc).probabilities(qargs)
                target_dict = Statevector(qc).probabilities_dict(qargs)
                Statevector(qc).probabilities_dict()
                self.assertTrue(np.allclose(probs, target))
                self.assertDictAlmostEqual(probs_dict, target_dict)
                StabilizerStateTestingTools._verify_individual_bitstrings(
                    self, target_dict, stab, qargs, dict_almost_equal=True
                )

    @combine(num_qubits=[2, 3, 4, 5])
    def test_expval_from_random_clifford(self, num_qubits):
        """Test that the expectation values for a random Clifford,
        where the Pauli operators are all its stabilizers,
        are equal to 1."""

        for _ in range(self.samples):
            cliff = random_clifford(num_qubits, seed=self.rng)
            qc = cliff.to_circuit()
            stab = StabilizerState(qc)
            stab_gen = stab.clifford.to_dict()["stabilizer"]
            for i in range(num_qubits):
                op = Pauli(stab_gen[i])
                exp_val = stab.expectation_value(op)
                self.assertEqual(exp_val, 1)

    def test_sample_counts_reset_bell(self):
        """Test sample_counts after reset for Bell state"""

        num_qubits = 2
        qc = QuantumCircuit(num_qubits)
        qc.h(0)
        qc.cx(0, 1)
        stab = StabilizerState(qc)

        target = {"00": self.shots / 2, "10": self.shots / 2}
        counts = {"00": 0, "10": 0}
        for _ in range(self.shots):
            res = stab.reset([0])
            value = res.measure()[0]
            counts[value] += 1
        self.assertDictAlmostEqual(counts, target, self.threshold)

        target = {"00": self.shots / 2, "01": self.shots / 2}
        counts = {"00": 0, "01": 0}
        for _ in range(self.shots):
            res = stab.reset([1])
            value = res.measure()[0]
            counts[value] += 1
        self.assertDictAlmostEqual(counts, target, self.threshold)

    def test_sample_counts_memory_ghz(self):
        """Test sample_counts and sample_memory method for GHZ state"""

        num_qubits = 3
        qc = QuantumCircuit(num_qubits)
        qc.h(0)
        qc.cx(0, 1)
        qc.cx(0, 2)
        stab = StabilizerState(qc)

        # 3-qubit qargs
        target = {"000": self.shots / 2, "111": self.shots / 2}
        for qargs in [[0, 1, 2], [2, 1, 0], [1, 2, 0], [1, 0, 2]]:
            with self.subTest(msg=f"counts (qargs={qargs})"):
                counts = stab.sample_counts(self.shots, qargs=qargs)
                self.assertDictAlmostEqual(counts, target, self.threshold)

            with self.subTest(msg=f"memory (qargs={qargs})"):
                memory = stab.sample_memory(self.shots, qargs=qargs)
                self.assertEqual(len(memory), self.shots)
                self.assertEqual(set(memory), set(target))

        # 2-qubit qargs
        target = {"00": self.shots / 2, "11": self.shots / 2}
        for qargs in [[0, 1], [2, 1], [1, 2], [1, 0]]:
            with self.subTest(msg=f"counts (qargs={qargs})"):
                counts = stab.sample_counts(self.shots, qargs=qargs)
                self.assertDictAlmostEqual(counts, target, self.threshold)

            with self.subTest(msg=f"memory (qargs={qargs})"):
                memory = stab.sample_memory(self.shots, qargs=qargs)
                self.assertEqual(len(memory), self.shots)
                self.assertEqual(set(memory), set(target))

        # 1-qubit qargs
        target = {"0": self.shots / 2, "1": self.shots / 2}
        for qargs in [[0], [1], [2]]:
            with self.subTest(msg=f"counts (qargs={qargs})"):
                counts = stab.sample_counts(self.shots, qargs=qargs)
                self.assertDictAlmostEqual(counts, target, self.threshold)

            with self.subTest(msg=f"memory (qargs={qargs})"):
                memory = stab.sample_memory(self.shots, qargs=qargs)
                self.assertEqual(len(memory), self.shots)
                self.assertEqual(set(memory), set(target))

    def test_sample_counts_memory_superposition(self):
        """Test sample_counts and sample_memory method of a 3-qubit superposition"""

        num_qubits = 3
        qc = QuantumCircuit(num_qubits)
        qc.h(0)
        qc.h(1)
        qc.h(2)
        stab = StabilizerState(qc)

        # 3-qubit qargs
        target = {
            "000": self.shots / 8,
            "001": self.shots / 8,
            "010": self.shots / 8,
            "011": self.shots / 8,
            "100": self.shots / 8,
            "101": self.shots / 8,
            "110": self.shots / 8,
            "111": self.shots / 8,
        }
        for qargs in [[0, 1, 2], [2, 1, 0], [1, 2, 0], [1, 0, 2]]:
            with self.subTest(msg=f"counts (qargs={qargs})"):
                counts = stab.sample_counts(self.shots, qargs=qargs)
                self.assertDictAlmostEqual(counts, target, self.threshold)

            with self.subTest(msg=f"memory (qargs={qargs})"):
                memory = stab.sample_memory(self.shots, qargs=qargs)
                self.assertEqual(len(memory), self.shots)
                self.assertEqual(set(memory), set(target))

        # 2-qubit qargs
        target = {
            "00": self.shots / 4,
            "01": self.shots / 4,
            "10": self.shots / 4,
            "11": self.shots / 4,
        }
        for qargs in [[0, 1], [2, 1], [1, 2], [1, 0]]:
            with self.subTest(msg=f"counts (qargs={qargs})"):
                counts = stab.sample_counts(self.shots, qargs=qargs)
                self.assertDictAlmostEqual(counts, target, self.threshold)

            with self.subTest(msg=f"memory (qargs={qargs})"):
                memory = stab.sample_memory(self.shots, qargs=qargs)
                self.assertEqual(len(memory), self.shots)
                self.assertEqual(set(memory), set(target))

        # 1-qubit qargs
        target = {"0": self.shots / 2, "1": self.shots / 2}
        for qargs in [[0], [1], [2]]:
            with self.subTest(msg=f"counts (qargs={qargs})"):
                counts = stab.sample_counts(self.shots, qargs=qargs)
                self.assertDictAlmostEqual(counts, target, self.threshold)

            with self.subTest(msg=f"memory (qargs={qargs})"):
                memory = stab.sample_memory(self.shots, qargs=qargs)
                self.assertEqual(len(memory), self.shots)
                self.assertEqual(set(memory), set(target))


@ddt
class TestStabilizerStateExpectationValue(QiskitTestCase):
    """Tests for StabilizerState.expectation_value method."""

    rng = np.random.default_rng(12345)
    samples = 10
    shots = 1000
    threshold = 0.1 * shots

    @data(("Z", 1), ("X", 0), ("Y", 0), ("I", 1), ("Z", 1), ("-Z", -1), ("iZ", 1j), ("-iZ", -1j))
    @unpack
    def test_expval_single_qubit_0(self, label, target):
        """Test expectation_value method of a single qubit on |0>"""
        qc = QuantumCircuit(1)
        stab = StabilizerState(qc)
        op = Pauli(label)
        expval = stab.expectation_value(op)
        self.assertEqual(expval, target)

    @data(("Z", -1), ("X", 0), ("Y", 0), ("I", 1))
    @unpack
    def test_expval_single_qubit_1(self, label, target):
        """Test expectation_value method of a single qubit on |1>"""
        qc = QuantumCircuit(1)
        qc.x(0)
        stab = StabilizerState(qc)
        op = Pauli(label)
        expval = stab.expectation_value(op)
        self.assertEqual(expval, target)

    @data(("Z", 0), ("X", 1), ("Y", 0), ("I", 1), ("X", 1), ("-X", -1), ("iX", 1j), ("-iX", -1j))
    @unpack
    def test_expval_single_qubit_plus(self, label, target):
        """Test expectation_value method of a single qubit on |+>"""
        qc = QuantumCircuit(1)
        qc.h(0)
        stab = StabilizerState(qc)
        op = Pauli(label)
        expval = stab.expectation_value(op)
        self.assertEqual(expval, target)

    @data(
        ("II", 1),
        ("XX", 0),
        ("YY", 0),
        ("ZZ", 1),
        ("IX", 0),
        ("IY", 0),
        ("IZ", 1),
        ("XY", 0),
        ("XZ", 0),
        ("YZ", 0),
        ("-ZZ", -1),
        ("iZZ", 1j),
        ("-iZZ", -1j),
    )
    @unpack
    def test_expval_two_qubits_00(self, label, target):
        """Test expectation_value method of two qubits in |00>"""

        num_qubits = 2

        qc = QuantumCircuit(num_qubits)
        stab = StabilizerState(qc)

        op = Pauli(label)
        expval = stab.expectation_value(op)
        self.assertEqual(expval, target)

    @data(
        ("II", 1),
        ("XX", 0),
        ("YY", 0),
        ("ZZ", 1),
        ("IX", 0),
        ("IY", 0),
        ("IZ", -1),
        ("XY", 0),
        ("XZ", 0),
        ("YZ", 0),
    )
    @unpack
    def test_expval_two_qubits_11(self, label, target):
        """Test expectation_value method of two qubits in |11>"""
        qc = QuantumCircuit(2)
        qc.x(0)
        qc.x(1)
        stab = StabilizerState(qc)
        op = Pauli(label)
        expval = stab.expectation_value(op)
        self.assertEqual(expval, target)

    @data(
        ("II", 1),
        ("XX", 1),
        ("YY", 0),
        ("ZZ", 0),
        ("IX", 1),
        ("IY", 0),
        ("IZ", 0),
        ("XY", 0),
        ("XZ", 0),
        ("YZ", 0),
        ("-XX", -1),
        ("iXX", 1j),
        ("-iXX", -1j),
    )
    @unpack
    def test_expval_two_qubits_plusplus(self, label, target):
        """Test expectation_value method of two qubits in |++>"""
        qc = QuantumCircuit(2)
        qc.h(0)
        qc.h(1)
        stab = StabilizerState(qc)

        op = Pauli(label)
        expval = stab.expectation_value(op)
        self.assertEqual(expval, target)

    @data(
        ("II", 1),
        ("XX", 0),
        ("YY", 0),
        ("ZZ", 0),
        ("IX", 0),
        ("IY", 0),
        ("IZ", -1),
        ("XY", 0),
        ("XZ", -1),
        ("YZ", 0),
    )
    @unpack
    def test_expval_two_qubits_plus1(self, label, target):
        """Test expectation_value method of two qubits in |+1>"""
        qc = QuantumCircuit(2)
        qc.x(0)
        qc.h(1)
        stab = StabilizerState(qc)

        op = Pauli(label)
        expval = stab.expectation_value(op)
        self.assertEqual(expval, target)

    @data(
        ("II", 1),
        ("XX", 1),
        ("YY", -1),
        ("ZZ", 1),
        ("IX", 0),
        ("IY", 0),
        ("IZ", 0),
        ("XY", 0),
        ("XZ", 0),
        ("YZ", 0),
        ("-YY", 1),
        ("iYY", -1j),
        ("-iYY", 1j),
    )
    @unpack
    def test_expval_two_qubits_bell_phi_plus(self, label, target):
        """Test expectation_value method of two qubits in bell phi plus"""
        qc = QuantumCircuit(2)
        qc.h(0)
        qc.cx(0, 1)
        stab = StabilizerState(qc)

        op = Pauli(label)
        expval = stab.expectation_value(op)
        self.assertEqual(expval, target)

    @data(
        ("II", 1),
        ("XX", 1),
        ("YY", 1),
        ("ZZ", -1),
        ("IX", 0),
        ("IY", 0),
        ("IZ", 0),
        ("XY", 0),
        ("XZ", 0),
        ("YZ", 0),
        ("-XX", -1),
        ("-YY", -1),
        ("iXX", 1j),
        ("iYY", 1j),
        ("-iXX", -1j),
        ("-iYY", -1j),
    )
    @unpack
    def test_expval_two_qubits_bell_phi_minus(self, label, target):
        """Test expectation_value method of two qubits in bell phi minus"""

        qc = QuantumCircuit(2)
        qc.h(0)
        qc.x(1)
        qc.cx(0, 1)
        stab = StabilizerState(qc)

        op = Pauli(label)
        expval = stab.expectation_value(op)
        self.assertEqual(expval, target)

    @data(
        ("II", 1),
        ("XX", 1),
        ("YY", 1),
        ("ZZ", -1),
        ("IX", 0),
        ("IY", 0),
        ("IZ", 0),
        ("XY", 0),
        ("XZ", 0),
        ("YZ", 0),
        ("-XX", -1),
        ("-YY", -1),
        ("iXX", 1j),
        ("iYY", 1j),
        ("-iXX", -1j),
        ("-iYY", -1j),
    )
    @unpack
    def test_expval_two_qubits_bell_sdg_h(self, label, target):
        """Test expectation_value method of two qubits in bell followed by sdg and h"""
        qc = QuantumCircuit(2)
        qc.h(0)
        qc.cx(0, 1)
        qc.sdg(0)
        qc.sdg(1)
        qc.h(0)
        qc.h(1)
        stab = StabilizerState(qc)

        op = Pauli(label)
        expval = stab.expectation_value(op)
        self.assertEqual(expval, target)

    @combine(num_qubits=[2, 3, 4, 5])
    def test_expval_random(self, num_qubits):
        """Test expectation_value method of random Cliffords"""

        for _ in range(self.samples):
            cliff = random_clifford(num_qubits, seed=self.rng)
            op = random_pauli(num_qubits, group_phase=True, seed=self.rng)
            qc = cliff.to_circuit()
            stab = StabilizerState(cliff)
            exp_val = stab.expectation_value(op)
            target = Statevector(qc).expectation_value(op)
            self.assertAlmostEqual(exp_val, target)

    @combine(num_qubits=[2, 3, 4, 5])
    def test_expval_random_subsystem(self, num_qubits):
        """Test expectation_value method of random Cliffords and a subsystem"""

        for _ in range(self.samples):
            cliff = random_clifford(num_qubits, seed=self.rng)
            op = random_pauli(2, group_phase=True, seed=self.rng)
            qargs = np.random.choice(num_qubits, size=2, replace=False)
            qc = cliff.to_circuit()
            stab = StabilizerState(cliff)
            exp_val = stab.expectation_value(op, qargs)
            target = Statevector(qc).expectation_value(op, qargs)
            self.assertAlmostEqual(exp_val, target)

    def test_stabilizer_bell_equiv(self):
        """Test that two circuits produce the same stabilizer group."""

        qc1 = QuantumCircuit(2)
        qc1.h(0)
        qc1.x(1)
        qc1.cx(0, 1)

        qc2 = QuantumCircuit(2)
        qc2.h(0)
        qc2.cx(0, 1)
        qc2.sdg(0)
        qc2.sdg(1)
        qc2.h(0)
        qc2.h(1)

        qc3 = QuantumCircuit(2)
        qc3.h(0)
        qc3.cx(0, 1)

        qc4 = QuantumCircuit(2)
        qc4.h(0)
        qc4.cx(0, 1)
        qc4.s(0)
        qc4.sdg(1)
        qc4.h(0)
        qc4.h(1)

        cliff1 = StabilizerState(qc1)  # ['+XX', '-ZZ']
        cliff2 = StabilizerState(qc2)  # ['+YY', '+XX']
        cliff3 = StabilizerState(qc3)  # ['+XX', '+ZZ']
        cliff4 = StabilizerState(qc4)  # ['-YY', '+XX']

        # [XX, -ZZ] and [XX, YY] both generate the stabilizer group {II, XX, YY, -ZZ}
        self.assertTrue(cliff1.equiv(cliff2))
        self.assertEqual(cliff1.probabilities_dict(), cliff2.probabilities_dict())
        StabilizerStateTestingTools._verify_individual_bitstrings(
            self, cliff1.probabilities_dict(), cliff2
        )
        StabilizerStateTestingTools._verify_individual_bitstrings(
            self, cliff2.probabilities_dict(), cliff1
        )

        # [XX, ZZ] and [XX, -YY] both generate the stabilizer group {II, XX, -YY, ZZ}
        self.assertTrue(cliff3.equiv(cliff4))
        self.assertEqual(cliff3.probabilities_dict(), cliff4.probabilities_dict())
        StabilizerStateTestingTools._verify_individual_bitstrings(
            self, cliff3.probabilities_dict(), cliff4
        )
        StabilizerStateTestingTools._verify_individual_bitstrings(
            self, cliff4.probabilities_dict(), cliff3
        )

        self.assertFalse(cliff1.equiv(cliff3))
        self.assertFalse(cliff2.equiv(cliff4))

    def test_visualize_does_not_throw_error(self):
        """Test to verify that drawing StabilizerState does not throw an error"""
        clifford = random_clifford(3, seed=0)
        stab = StabilizerState(clifford)
        _ = repr(stab)


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=no-member,invalid-name,missing-docstring,no-name-in-module
# pylint: disable=attribute-defined-outside-init,unsubscriptable-object

from qiskit import converters

from .utils import random_circuit


class ConverterBenchmarks:
    params = ([1, 2, 5, 8, 14, 20, 32, 53], [8, 128, 2048, 8192])
    param_names = ["n_qubits", "depth"]
    timeout = 600

    def setup(self, n_qubits, depth):
        seed = 42
        # NOTE: Remove the benchmarks larger than 20x2048 and 14x8192, this is
        # a tradeoff for speed of benchmarking, creating circuits this size
        # takes more time than is worth it for benchmarks that take a couple
        # seconds
        if n_qubits >= 20:
            if depth >= 2048:
                raise NotImplementedError
        elif n_qubits == 14:
            if depth > 2048:
                raise NotImplementedError
        self.qc = random_circuit(n_qubits, depth, measure=True, conditional=True, seed=seed)
        self.dag = converters.circuit_to_dag(self.qc)

    def time_circuit_to_dag(self, *_):
        converters.circuit_to_dag(self.qc)

    def time_circuit_to_instruction(self, *_):
        converters.circuit_to_instruction(self.qc)

    def time_dag_to_circuit(self, *_):
        converters.dag_to_circuit(self.dag)
